import os
import sys
import shutil
from subprocess import call
from annogesiclib.helper import Helper
from annogesiclib.multiparser import Multiparser
from annogesiclib.converter import Converter
from annogesiclib.merge_manual import merge_manual_predict_tss
from annogesiclib.stat_TSSpredator import stat_tsspredator
from annogesiclib.plot_TSS_venn import plot_venn
from annogesiclib.validate_gene import validate_gff
from annogesiclib.stat_TA_comparison import stat_ta_tss
from annogesiclib.check_orphan import check_orphan
from annogesiclib.filter_TSS_pro import filter_tss_pro
from annogesiclib.filter_low_expression import filter_low_expression


class TSSpredator(object):

    def __init__(self, args_tss):
        self.multiparser = Multiparser()
        self.helper = Helper()
        self.converter = Converter()
        self.master = os.path.join(args_tss.out_folder, "MasterTables")
        self.tmps = {"tss": "tmp_TSS", "ta_tss": "tmp_ta_tss", "tss_ta":
                     "tmp_tss", "tmp": "tmp"}
        if args_tss.ta_files is not None:
            self.tmps["ta"] = os.path.join(args_tss.ta_files, "tmp")
        else:
            self.tmps["ta"] = None
        self.gff_path = os.path.join(args_tss.gffs, "tmp")
        if args_tss.manual is not None:
            self.manual_path = os.path.join(args_tss.manual, "tmp")
        self.wig_path = os.path.join(args_tss.wig_folder, "tmp")
        self.fasta_path = os.path.join(args_tss.fastas, "tmp")
        self.stat_outfolder = os.path.join(args_tss.out_folder, "statistics")
        self.gff_outfolder = os.path.join(args_tss.out_folder, "gffs")

    def _assign_dict(self, lib_datas):
        return {"wig": lib_datas[0],
                "tex": lib_datas[1],
                "condition": int(lib_datas[2]),
                "replicate": lib_datas[3],
                "strand": lib_datas[4]}

    def _print_lib(self, lib_num, lib_list, out, wig_folder, prefix, rep_set):
        for num_id in range(1, lib_num+1):
            cond_list = []
            for lib in lib_list:
                if num_id == lib["condition"]:
                    cond_list.append(lib)
            cond_sort_list = sorted(cond_list, key=lambda k: k['replicate'])
            reps = []
            for cond in cond_sort_list:
                out.write("{0}_{1}{2} = {3}\n".format(
                          prefix, cond["condition"], cond["replicate"],
                          os.path.join(wig_folder, cond["wig"])))
                reps.append(cond["replicate"])
            for rep in sorted(rep_set):
                if rep not in reps:
                    out.write("{0}_{1}{2} = \n".format(
                              prefix, cond["condition"], rep))

    def _start_to_run(self, tsspredator_path, config_file, out_path, prefix):
        print("Running TSSpredator for " + prefix)
        out = open(os.path.join(out_path, "log.txt"), "w")
        err = open(os.path.join(out_path, "err.txt"), "w")
        call(["java", "-jar", tsspredator_path,
              config_file], stdout=out, stderr=err)
        out.close()
        err.close()

    def _import_lib(self, libs, wig_folder, project_strain_name,
                    out, gff, program, fasta):
        lib_dict = {"fp": [], "fm": [], "nm": [], "np": []}
        lib_num = 0
        rep_set = set()
        list_num_id = []
        for lib in libs:
            lib_datas = lib.split(":")
            if not lib_datas[0].endswith(".wig"):
                print("Error: Wiggle files are not end with .wig!")
                sys.exit()
            for wig in os.listdir(wig_folder):
                filename = wig.split("_STRAIN_")
                if (filename[0] == lib_datas[0][:-4]) and (
                        filename[1][:-4] == project_strain_name):
                    lib_datas[0] = wig
            if int(lib_datas[2]) > lib_num:
                lib_num = int(lib_datas[2])
            if lib_datas[3] not in rep_set:
                rep_set.add(lib_datas[3])
            if (lib_datas[1] == "tex") and (lib_datas[4] == "+"):
                lib_dict["fp"].append(self._assign_dict(lib_datas))
            elif (lib_datas[1] == "tex") and (lib_datas[4] == "-"):
                lib_dict["fm"].append(self._assign_dict(lib_datas))
            elif (lib_datas[1] == "notex") and (lib_datas[4] == "+"):
                lib_dict["np"].append(self._assign_dict(lib_datas))
            elif (lib_datas[1] == "notex") and (lib_datas[4] == "-"):
                lib_dict["nm"].append(self._assign_dict(lib_datas))
        for num_id in range(1, lib_num+1):
            out.write("annotation_{0} = {1}\n".format(num_id, gff))
        if program.lower() == "tss":
            self._print_lib(lib_num, lib_dict["fm"], out,
                            wig_folder, "fivePrimeMinus", rep_set)
            self._print_lib(lib_num, lib_dict["fp"], out,
                            wig_folder, "fivePrimePlus", rep_set)
        elif program.lower() == "ps":
            self._print_lib(lib_num, lib_dict["nm"], out,
                            wig_folder, "fivePrimeMinus", rep_set)
            self._print_lib(lib_num, lib_dict["np"], out,
                            wig_folder, "fivePrimePlus", rep_set)
        else:
            print("Error: Wrong program name! Please assing tss "
                  "or processing_site.")
            sys.exit()
        for num_id in range(1, lib_num+1):
            out.write("genome_{0} = {1}\n".format(num_id, fasta))
        for num_id in range(1, lib_num+1):
            list_num_id.append(str(num_id))
        return lib_num, num_id, rep_set, lib_dict, list_num_id

    def _print_repmatch(self, args_tss, out):
        '''check replicate match'''
        detect_all = False
        for rep in args_tss.repmatch:
            if "all" in rep:
                detect_all = True
                match = rep.split("_")[-1]
                out.write("minNumRepMatches = {0}\n".format(match))
                break
        if not detect_all:
            nums = {}
            matchs = {}
            for match in args_tss.repmatch:
                lib = match.split("_")[0]
                rep = match.split("_")[-1]
                matchs[lib] = rep
                if rep not in nums.keys():
                    nums[rep] = 1
                else:
                    nums[rep] += 1
            for rep, num in nums.items():
                if num == max(nums.values()):
                    out.write("minNumRepMatches = {0}\n".format(rep))
                    max_rep = rep
                    break
            for lib, rep in matchs.items():
                if rep != max_rep:
                    out.write("minNumRepMatches_{0} = {1}\n".format(
                        lib, rep))

    def _gen_config(self, project_strain_name, args_tss, gff,
                    wig_folder, fasta, config_file):
        '''generation of config files'''
        master_folder = "MasterTable_" + project_strain_name
        out_path = os.path.join(self.master, master_folder)
        self.helper.check_make_folder(out_path)
        out = open(config_file, "w")
        out.write("TSSinClusterSelectionMethod = HIGHEST\n")
        out.write("allowedCompareShift = 1\n")
        out.write("allowedRepCompareShift = 1\n")
        lib_num, num_id, rep_set, lib_dict, list_num_id = \
            self._import_lib(args_tss.libs, wig_folder, project_strain_name,
                             out, gff, args_tss.program, fasta)
        out.write("idList = ")
        out.write(",".join(list_num_id) + "\n")
        out.write("maxASutrLength = 100\n")
        out.write("maxGapLengthInGene = 500\n")
        out.write("maxNormalTo5primeFactor = {0}\n".format(
                  args_tss.processing_factor))
        out.write("maxTSSinClusterDistance = {0}\n".format(
                  args_tss.cluster + 1))
        out.write("maxUTRlength = {0}\n".format(args_tss.utr_length))
        out.write("min5primeToNormalFactor = {0}\n".format(
                  args_tss.enrichment_factor))
        out.write("minCliffFactor = {0}\n".format(args_tss.factor))
        out.write("minCliffFactorDiscount = {0}\n".format(
                  args_tss.factor_reduction))
        out.write("minCliffHeight = {0}\n".format(args_tss.height))
        out.write("minCliffHeightDiscount = {0}\n".format(
                  args_tss.height_reduction))
        out.write("minNormalHeight = {0}\n".format(args_tss.base_height))
        self._print_repmatch(args_tss, out)
        out.write("minPlateauLength = 0\n")
        out.write("mode = cond\n")
        out.write("normPercentile = 0.9\n")
        if args_tss.program.lower() == "tss":
            self._print_lib(lib_num, lib_dict["nm"], out,
                            wig_folder, "normalMinus", rep_set)
            self._print_lib(lib_num, lib_dict["np"], out,
                            wig_folder, "normalPlus", rep_set)
        else:
            self._print_lib(lib_num, lib_dict["fm"], out,
                            wig_folder, "normalMinus", rep_set)
            self._print_lib(lib_num, lib_dict["fp"], out,
                            wig_folder, "normalPlus", rep_set)
        out.write("numReplicates = {0}\n".format(len(rep_set)))
        out.write("numberOfDatasets = {0}\n".format(lib_num))
        out.write("outputDirectory = {0}\n".format(out_path))
        for prefix_id in range(len(args_tss.output_prefixs)):
            out.write("outputPrefix_{0} = {1}\n".format(
                      prefix_id + 1, args_tss.output_prefixs[prefix_id]))
        out.write("projectName = {0}\n".format(project_strain_name))
        out.write("superGraphCompatibility = igb\n")
        out.write("texNormPercentile = 0.5\n")
        out.write("writeGraphs = 0\n")
        out.write("writeNocornacFiles = 0\n")
        out.close()

    def _convert_gff(self, prefixs, args_tss):
        for prefix in prefixs:
            out_file = os.path.join(self.gff_outfolder, "_".join([
                           prefix, args_tss.program]) + ".gff")
            gff_f = open(out_file, "w")
            out_path = os.path.join(self.master, "_".join([
                           "MasterTable", prefix]))
            if "MasterTable.tsv" not in os.listdir(out_path):
                print("Error: There is not MasterTable file in {0}".format(
                      out_path))
                print("Please check configuration file.")
            else:
                if args_tss.program.lower() == "processing":
                    feature = "processing_site"
                elif args_tss.program.lower() == "tss":
                    feature = "TSS"
                self.converter.convert_mastertable2gff(
                    os.path.join(out_path, "MasterTable.tsv"),
                    "ANNOgesic", feature, prefix, out_file)
            gff_f.close()

    def _merge_manual(self, tsss, args_tss):
        '''if manual detected TSS is provided, it can merge manual detected TSS 
        and TSSpredator predicted TSS'''
        self.helper.check_make_folder(os.path.join(os.getcwd(),
                                      self.tmps["tss"]))
        for tss in tsss:
            for gff in os.listdir(args_tss.gffs):
                if (gff[:-4] == tss) and (".gff" in gff):
                    break
            filename = "_".join([tss, args_tss.program]) + ".gff"
            predict = os.path.join(self.gff_outfolder, filename)
            manual = os.path.join(self.manual_path, tss + ".gff")
            fasta = os.path.join(self.fasta_path, tss + ".fa")
            stat_file = "stat_compare_TSSpredator_manual_{0}.csv".format(tss)
            if os.path.exists(manual):
                print("Merging and classiflying manual-detected "
                      "TSSs for {0}".format(tss))
                merge_manual_predict_tss(
                    predict, stat_file,
                    os.path.join(self.tmps["tss"], filename),
                    os.path.join(args_tss.gffs, gff), args_tss, manual, fasta)
            if os.path.exists(stat_file):
                shutil.move(stat_file, os.path.join(
                    args_tss.out_folder, "statistics", tss, stat_file))
        self.helper.move_all_content(self.tmps["tss"],
                                     self.gff_outfolder, [".gff"])
        shutil.rmtree(self.tmps["tss"])

    def _validate(self, tsss, args_tss):
        '''validate TSS with genome annotation'''
        print("Validating TSSs with genome annotations")
        for tss in tsss:
            for gff in os.listdir(args_tss.gffs):
                if (gff[:-4] == tss) and (".gff" in gff):
                    break
            stat_file = os.path.join(
                    self.stat_outfolder, tss,
                    "".join(["stat_gene_vali_", tss, ".csv"]))
            out_cds_file = os.path.join(args_tss.out_folder, "tmp.gff")
            if args_tss.program.lower() == "tss":
                compare_file = os.path.join(self.gff_outfolder,
                                            "_".join([tss, "TSS.gff"]))
            elif args_tss.program.lower() == "processing":
                compare_file = os.path.join(self.gff_outfolder,
                                            "_".join([tss, "processing.gff"]))
            validate_gff(compare_file, os.path.join(args_tss.gffs, gff),
                         stat_file, out_cds_file, args_tss.utr_length,
                         args_tss.program.lower())
            shutil.move(out_cds_file, os.path.join(args_tss.gffs, gff))

    def _compare_ta(self, tsss, args_tss):
        '''compare TSS with transcript'''
        detect = False
        print("Comparing transcripts and TSSs")
        self.multiparser.parser_gff(args_tss.ta_files, "transcript")
        self.multiparser.combine_gff(args_tss.gffs, self.tmps["ta"],
                                     None, "transcript")
        for tss in tsss:
            stat_out = os.path.join(
                    self.stat_outfolder, tss, "".join([
                        "stat_compare_TSS_transcript_",
                        tss, ".csv"]))
            for ta in os.listdir(self.tmps["ta"]):
                filename = ta.split("_transcript")
                if (filename[0] == tss) and (filename[1] == ".gff"):
                    detect = True
                    break
            compare_file = os.path.join(self.gff_outfolder,
                                        "_".join([tss, "TSS.gff"]))
            if detect:
                stat_ta_tss(os.path.join(self.tmps["ta"], ta), compare_file,
                            stat_out, self.tmps["ta_tss"],
                            self.tmps["tss_ta"], args_tss.fuzzy)
                self.helper.sort_gff(self.tmps["tss_ta"], compare_file)
                self.helper.sort_gff(self.tmps["ta_tss"],
                                     os.path.join(args_tss.ta_files, ta))
                os.remove(self.tmps["tss_ta"])
                os.remove(self.tmps["ta_tss"])
                detect = False

    def _stat_tss(self, tsss, feature):
        print("Running statistaics")
        for tss in tsss:
            compare_file = os.path.join(self.gff_outfolder,
                                        "_".join([tss, feature]) + ".gff")
            stat_tsspredator(
                compare_file, feature,
                os.path.join(self.stat_outfolder, tss, "_".join([
                    "stat", feature, "class", tss]) + ".csv"),
                os.path.join(self.stat_outfolder, tss, "_".join([
                    "stat", feature, "libs", tss]) + ".csv"))
            self.helper.move_all_content(os.getcwd(), os.path.join(
                self.stat_outfolder, tss), ["_class", ".png"])
            if os.path.exists(os.path.join(
                    self.stat_outfolder, "TSSstatistics.tsv")):
                shutil.move(
                    os.path.join(
                        self.stat_outfolder, "TSSstatistics.tsv"),
                    os.path.join(
                        self.stat_outfolder, tss, "TSSstatistics.tsv"))
            plot_venn(compare_file, feature)
            self.helper.move_all_content(os.getcwd(), os.path.join(
                self.stat_outfolder, tss), ["_venn", ".png"])

    def _set_gen_config(self, args_tss, input_folder):
        prefixs = []
        detect = False
        for fasta in os.listdir(self.fasta_path):
            run = False
            for gff in os.listdir(self.gff_path):
                if fasta[:-3] == gff[:-4]:
                    prefix = fasta[:-3]
                    for wig in os.listdir(self.wig_path):
                        filename = wig.split("_STRAIN_")
                        if filename[1][:-4] == prefix:
                            detect = True
                            break
                    if detect:
                        prefixs.append(prefix)
                        config = os.path.join(
                                input_folder,
                                "_".join(["config", prefix]) + ".ini")
                        self._gen_config(
                            prefix, args_tss,
                            os.path.join(self.gff_path, gff), self.wig_path,
                            os.path.join(self.fasta_path, fasta), config)
        return prefixs

    def _merge_wigs(self, wig_folder, prefix, libs):
        self.helper.check_make_folder(os.path.join(os.getcwd(),
                                      self.tmps["tmp"]))
        for wig_file in os.listdir(wig_folder):
            for lib in libs:
                info = lib.split(":")
                if (info[0][:-4] in wig_file) and (info[-1] == "+") and (
                        prefix in wig_file) and (
                        os.path.isfile(os.path.join(wig_folder, wig_file))):
                    Helper().merge_file(
                            os.path.join(wig_folder, wig_file),
                            os.path.join("tmp", "merge_forward.wig"))
                if (info[0][:-4] in wig_file) and (info[-1] == "-") and (
                        prefix in wig_file) and (
                        os.path.isfile(os.path.join(wig_folder, wig_file))):
                    Helper().merge_file(
                            os.path.join(wig_folder, wig_file),
                            os.path.join("tmp", "merge_reverse.wig"))

    def _check_orphan(self, prefixs, wig_folder, args_tss):
        '''if genome has no locus tag, it can use for classify the TSS'''
        for prefix in prefixs:
            self._merge_wigs(wig_folder, prefix, args_tss.libs)
            tmp_tss = os.path.join(self.tmps["tmp"], "_".join([
                          prefix, args_tss.program + ".gff"]))
            pre_tss = os.path.join(self.gff_outfolder, "_".join([
                          prefix, args_tss.program + ".gff"]))
            check_orphan(pre_tss, os.path.join(
                args_tss.gffs, prefix + ".gff"),
                "tmp/merge_forward.wig", "tmp/merge_reverse.wig", tmp_tss)
            shutil.move(tmp_tss, pre_tss)
        shutil.rmtree("tmp")

    def _remove_files(self, args_tss):
        print("Remove temperary files and folders")
        self.helper.remove_tmp_dir(args_tss.fastas)
        self.helper.remove_tmp_dir(args_tss.gffs)
        self.helper.remove_tmp_dir(args_tss.ta_files)
        if "merge_forward.wig" in os.listdir(os.getcwd()):
            os.remove("merge_forward.wig")
        if "merge_reverse.wig" in os.listdir(os.getcwd()):
            os.remove("merge_reverse.wig")
        shutil.rmtree(args_tss.wig_folder)
        if args_tss.manual is not None:
            shutil.rmtree(args_tss.manual)

    def _deal_with_overlap(self, out_folder, args_tss):
        '''deal with the situation that TSS and 
        processing site at the same position'''
        if not args_tss.overlap_feature:
            pass
        else:
            print("Comparing TSSs and Processing sites")
            if args_tss.program.lower() == "tss":
                for tss in os.listdir(out_folder):
                    if tss.endswith("_TSS.gff"):
                        ref = self.helper.get_correct_file(
                                args_tss.overlap_gffs, "_processing.gff",
                                tss.replace("_TSS.gff", ""), None, None)
                        filter_tss_pro(os.path.join(out_folder, tss),
                                       ref, args_tss.program,
                                       args_tss.cluster)
            elif args_tss.program.lower() == "processing":
                for tss in os.listdir(out_folder):
                    if tss.endswith("_processing.gff"):
                        ref = self.helper.get_correct_file(
                                args_tss.overlap_gffs, "_TSS.gff",
                                tss.replace("_processing.gff", ""), None, None)
                        filter_tss_pro(os.path.join(out_folder, tss),
                                       ref, args_tss.program,
                                       args_tss.cluster)

    def _low_expression(self, args_tss, gff_folder):
        '''deal with the low expressed TSS'''
        prefix = None
        self._merge_wigs(args_tss.wig_folder, "wig", args_tss.libs)
        for gff in os.listdir(gff_folder):
            if (args_tss.program.lower() == "tss") and (
                    gff.endswith("_TSS.gff")):
                prefix = gff.replace("_TSS.gff", "")
            elif (args_tss.program.lower() == "processing") and (
                    gff.endswith("_processing.gff")):
                prefix = gff.replace("_processing.gff", "")
            if prefix:
                out = open(os.path.join(
                    self.stat_outfolder, prefix, "_".join([
                        "stat", prefix, "low_expression_cutoff.csv"])), "w")
                out.write("\t".join(["Genome", "Cutoff_coverage"]) + "\n")
                cutoff = filter_low_expression(
                        os.path.join(gff_folder, gff), args_tss,
                        "tmp/merge_forward.wig", "tmp/merge_reverse.wig",
                        "tmp/without_low_expression.gff")
                out.write("\t".join([prefix, str(cutoff)]) + "\n")
                os.remove(os.path.join(gff_folder, gff))
                shutil.move("tmp/without_low_expression.gff",
                            os.path.join(gff_folder, gff))
                prefix = None
        out.close()

    def run_tsspredator(self, args_tss):
        input_folder = os.path.join(args_tss.out_folder, "configs")
        for gff in os.listdir(args_tss.gffs):
            if gff.endswith(".gff"):
                self.helper.check_uni_attributes(os.path.join(
                                                 args_tss.gffs, gff))
        self.helper.check_make_folder(self.gff_outfolder)
        self.multiparser.parser_fasta(args_tss.fastas)
        self.multiparser.parser_gff(args_tss.gffs, None)
        self.multiparser.parser_wig(args_tss.wig_folder)
        prefixs = self._set_gen_config(args_tss, input_folder)
        for prefix in prefixs:
            out_path = os.path.join(
                    self.master, "_".join(["MasterTable", prefix]))
            config_file = os.path.join(
                    input_folder, "_".join(["config", prefix]) + ".ini")
            self._start_to_run(args_tss.tsspredator_path, config_file,
                               out_path, prefix)
            if os.path.exists(os.path.join(out_path, "TSSstatistics.tsv")):
                shutil.move(os.path.join(out_path, "TSSstatistics.tsv"),
                            os.path.join(
                                self.stat_outfolder, "TSSstatistics.tsv"))
        if args_tss.program.lower() == "ps":
            args_tss.program = "processing"
        self._convert_gff(prefixs, args_tss)
        if args_tss.check_orphan:
            print("checking the orphan TSSs")
            self._check_orphan(prefixs,
                               os.path.join(args_tss.wig_folder, "tmp"),
                               args_tss)
        self.multiparser.combine_gff(args_tss.gffs, self.gff_outfolder,
                                     None, args_tss.program)
        datas = []
        for gff in os.listdir(self.gff_outfolder):
            if gff.endswith(".gff"):
                gff_folder = gff.replace("".join(["_", args_tss.program,
                                                  ".gff"]), "")
                self.helper.check_make_folder(
                     os.path.join(self.stat_outfolder, gff_folder))
                datas.append(gff_folder)
        if args_tss.remove_low_expression is not None:
            self._low_expression(args_tss, self.gff_outfolder)
        if args_tss.manual is not None:
            self.multiparser.parser_gff(args_tss.manual, None)
            self.multiparser.combine_gff(args_tss.gffs, self.manual_path,
                                         None, None)
            self.multiparser.combine_fasta(args_tss.gffs, self.fasta_path,
                                         None)
            self.multiparser.combine_wig(args_tss.gffs, self.wig_path,
                                         None, args_tss.libs)
            self._merge_manual(datas, args_tss)
        self._deal_with_overlap(self.gff_outfolder, args_tss)
        self._stat_tss(datas, args_tss.program)
        if args_tss.validate:
            self._validate(datas, args_tss)
        if args_tss.ta_files is not None:
            self._compare_ta(datas, args_tss)
        self._remove_files(args_tss)



import os
import shutil
import csv
from annogesiclib.multiparser import Multiparser
from annogesiclib.seq_editer import SeqEditer
from annogesiclib.helper import Helper


class TargetFasta(object):
    '''detection of sRNA target interaction'''

    def __init__(self, tar_folder, ref_folder):
        self.multiparser = Multiparser()
        self.seq_editer = SeqEditer()
        self.helper = Helper()
        self.folders = {"tmp_tar": os.path.join(tar_folder, "tmp")}

    def gen_folder(self, out_folder, ref_files):
        new_ref_folder = os.path.join(out_folder, "tmp_reference")
        self.helper.check_make_folder(new_ref_folder)
        for file_ in ref_files:
            shutil.copy(file_, new_ref_folder)
        self.folders["tmp_ref"] = os.path.join(new_ref_folder, "tmp")
        self.multiparser.parser_fasta(new_ref_folder)
        if os.path.exists(self.folders["tmp_tar"]):
            shutil.rmtree(self.folders["tmp_tar"])
        os.mkdir(self.folders["tmp_tar"])
        return new_ref_folder

    def get_target_fasta(self, mut_table, tar_folder, ref_files,
                         combine, out_folder):
        pass
        new_ref_folder = self.gen_folder(out_folder, ref_files)
        self.seq_editer.modify_seq(self.folders["tmp_ref"], mut_table,
                                   self.folders["tmp_tar"])
        print("Transfering to query sequences")
        mh = open(mut_table, "r")
        pre_strain = None
        out = None
        for row in csv.reader(mh, delimiter='\t'):
            strain = row[1]
            if not row[0].startswith("#"):
                if (pre_strain != row[1]):
                    fasta = os.path.join(out_folder, "fasta_files",
                                         strain + ".fa")
                    if out is not None:
                        out.close()
                    out = open(fasta, "w")
                    if strain + ".fa" in os.listdir(self.folders["tmp_tar"]):
                        with open(os.path.join(
                                  self.folders["tmp_tar"],
                                  strain + ".fa")) as f_h:
                            for line in f_h:
                                out.write(line)
                    else:
                        print("Error: No fasta information of {0}.fa".format(
                              strain))
        out.close()
        if combine:
            out_seq = "updated_genomes.fa"
            if os.path.exists(out_seq):
                os.remove(out_seq)
            for seq in os.listdir(os.path.join(out_folder, "fasta_files")):
                if seq.endswith(".fa"):
                    os.system(" ".join(["cat", os.path.join(
                                            out_folder, "fasta_files", seq),
                                        ">>", out_seq]))
                    os.remove(os.path.join(out_folder, "fasta_files", seq))
            shutil.move(out_seq, os.path.join(
                out_folder, "fasta_files", out_seq))
        shutil.rmtree(self.folders["tmp_tar"])
        shutil.rmtree(self.folders["tmp_ref"])
        if "tmp_reference" in os.listdir(out_folder):
            shutil.rmtree(new_ref_folder)
        print("Please use the new fasta files to remapping again.")



import sys
import os
import gc
import numpy as np
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.coverage_detection import coverage_comparison
from annogesiclib.coverage_detection import replicate_comparison, get_repmatch
from annogesiclib.lib_reader import read_wig, read_libs
from annogesiclib.gen_TSS_type import compare_tss_cds, fix_primary_type
from annogesiclib.args_container import ArgsContainer


def get_differential_cover(num, checks, cover_sets, poss, cover,
                           args_srna, cover_pos):
    go_out = False
    if checks["detect_diff"]:
        if (num == args_srna.fuzzy_inter) or (
            cover_sets["diff"] == 0) or (
            (cover > cover_sets["diff"]) and (
             cover / cover_sets["diff"]) > (
             1 + args_srna.decrease_inter)):
            poss["stop_point"] = cover_pos
            go_out = True
        elif (cover <= cover_sets["diff"]):
            if (cover / cover_sets["diff"]) <= (
                    args_srna.decrease_inter / 2):
                num += 1
            else:
                num = 0
            cover_sets["diff"] = cover
            cover_sets["low"] = cover
        elif (cover > cover_sets["diff"]) and (
                (cover / cover_sets["diff"]) <= (
                1 + args_srna.decrease_inter)):
            num += 1
    if (not checks["first"]) and (cover_sets["high"] > 0):
        if ((cover_sets["low"] / cover_sets["high"]) <
                args_srna.decrease_inter) and (
                cover_sets["low"] > -1):
            checks["detect_diff"] = True
            cover_sets["diff"] = cover
    return go_out


def check_coverage_pos(start, end, cover, cutoff_coverage, cover_sets,
                       checks, poss, strand, cover_pos):
    go_out = False
    if (start <= cover_pos) and (
            end >= cover_pos):
        if cover > cutoff_coverage:
            cover_sets["total"] = cover_sets["total"] + cover
            checks["first"] = coverage_comparison(
                              cover, cover_sets, poss,
                              checks["first"], strand, cover_pos)
        else:
            cover_sets["total"] = cover_sets["total"] + cover
            checks["first"] = coverage_comparison(
                                cover, cover_sets, poss,
                                checks["first"], strand, cover_pos)
    else:
        if (strand == "+") and (cover_pos > end):
            poss["stop_point"] = cover_pos
            go_out = True
        elif (strand == "-") and (cover_pos < start):
            poss["stop_point"] = cover_pos
            go_out = True
    return go_out


def check_start_and_end(start, end, covers):
    if (start - 2) < 0:
        c_start = 0
    else:
        c_start = start - 2
    if (end + 2) > len(covers):
        c_end = len(covers)
    else:
        c_end = end + 2
    return c_start, c_end


def get_best(wigs, strain, strand, start, end, type_, args_srna, cutoff):
    cover_sets = {"low": -1, "high": -1, "total": 0, "diff": 0}
    poss = {"high": 0, "low": 0, "stop_point": -1}
    srna_covers = {}
    for wig_strain, conds in wigs.items():
        if wig_strain == strain:
            for cond, tracks in conds.items():
                srna_covers[cond] = []
                for lib_name, covers in tracks.items():
                    track = lib_name.split("|")[-3]
                    lib_strand = lib_name.split("|")[-2]
                    lib_type = lib_name.split("|")[-1]
                    cover_sets["total"] = 0
                    cover_sets["diff"] = 0
                    checks = {"first": True, "detect_diff": False}
                    num = 0
                    c_start, c_end = check_start_and_end(start, end, covers)
                    covers = covers[c_start: c_end]
                    if strand == "-":
                        covers = covers[::-1]
                    go_out = False
                    pos = 0
                    for cover in covers:
                        if strand == "+":
                            cover_pos = c_start + pos
                        else:
                            cover_pos = c_end - pos
                        if (lib_strand == strand):
                            go_out = check_coverage_pos(
                                    start, end, cover, cutoff,
                                    cover_sets, checks, poss,
                                    strand, cover_pos)
                            if go_out:
                                break
                            if type_ == "differential":
                                go_out = get_differential_cover(
                                        num, checks, cover_sets,
                                        poss, cover, args_srna, cover_pos)
                                if go_out:
                                    break
                        pos += 1
                    if strand == "+":
                        diff = poss["stop_point"] - start
                    else:
                        diff = end - poss["stop_point"]
                    avg = cover_sets["total"] / float(diff + 1)
                    if (avg > float(cutoff)):
                        srna_covers[cond].append({"track": track,
                                                  "high": cover_sets["high"],
                                                  "low": cover_sets["low"],
                                                  "avg": avg,
                                                  "pos": poss["stop_point"],
                                                  "type": lib_type})
    return srna_covers


def get_attribute_string(srna_datas, tss_pro, num, name, srna_type, strain):
    attribute_string = ";".join(
                    ["=".join(items) for items in (
                     ["ID", strain + "_srna" + str(num)],
                     ["Name", "_".join(["sRNA", name])],
                     ["sRNA_type", srna_type])])
    datas = tss_pro.split(";")
    tss = ""
    pro = ""
    for data in datas:
        if "TSS" in data:
            if len(tss) == 0:
                tss = tss + data
            else:
                tss = ";".join([tss, data])
        elif "Cleavage" in data:
            if len(pro) == 0:
                pro = pro + data
            else:
                pro = ";".join([pro, data])
    if len(tss) == 0:
        tss = "NA"
    if len(pro) == 0:
        pro = "NA"
    with_tss = "=".join(["with_TSS", tss])
    with_pro = "=".join(["end_cleavage", pro])
    if srna_datas is None:
        if (tss != "NA") and (pro != "NA"):
            attribute_string = ";".join([attribute_string, with_tss, with_pro])
        elif (tss != "NA"):
            attribute_string = ";".join([attribute_string, with_tss])
        elif (pro != "NA"):
            attribute_string = ";".join([attribute_string, with_pro])
    else:
        srna_data_string = ";".join(
                ["=".join(items) for items in (
                 ["best_avg_coverage", str(srna_datas["best"])],
                 ["best_high_coverage", str(srna_datas["high"])],
                 ["best_low_coverage", str(srna_datas["low"])])])
        if (tss != "NA") and (pro != "NA"):
            attribute_string = ";".join([attribute_string, with_tss, with_pro,
                                         srna_data_string])
        elif (tss != "NA"):
            attribute_string = ";".join([attribute_string, with_tss,
                                         srna_data_string])
        elif (pro != "NA"):
            attribute_string = ";".join([attribute_string, with_pro,
                                         srna_data_string])
        else:
            attribute_string = ";".join([attribute_string, srna_data_string])
    return attribute_string


def print_file(string, tss, srna_datas, srna_type, args_srna, strain):
    name = '%0*d' % (5, args_srna.nums["uni"])
    datas = string.split("\t")
    if (srna_datas is None):
        args_srna.out_table.write(
            "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t".format(
                datas[0], name, datas[3], datas[4], datas[6],
                "NA", "NA", "NA", "NA", "NA"))
        attribute_string = get_attribute_string(
                srna_datas, tss, args_srna.nums["uni"], name, srna_type,
                strain)
        args_srna.output.write("\t".join([string, attribute_string]) + "\n")
        args_srna.out_table.write(tss + "\n")
    else:
        args_srna.out_table.write(
            "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t".format(
                datas[0], name, datas[3], datas[4], datas[6],
                ";".join(srna_datas["conds"].keys()),
                ";".join(srna_datas["conds"].values()),
                srna_datas["best"], srna_datas["high"], srna_datas["low"]))
        attribute_string = get_attribute_string(
                srna_datas, tss, args_srna.nums["uni"], name, srna_type,
                strain)
        args_srna.output.write("\t".join([string, attribute_string]) + "\n")
        if srna_datas["detail"] is not None:
            args_srna.out_table.write(tss + "\t")
            if not args_srna.table_best:
                first = True
                for data in srna_datas["detail"]:
                    if first:
                        args_srna.out_table.write(
                            "{0}(avg={1};high={2};low={3})".format(
                                data["track"], data["avg"],
                                data["high"], data["low"]))
                        first = False
                    else:
                        args_srna.out_table.write(
                            ";{0}(avg={1};high={2};low={3})".format(
                                data["track"], data["avg"],
                                data["high"], data["low"]))
            else:
                args_srna.out_table.write(
                    "{0}(avg={1};high={2};low={3})".format(
                        srna_datas["track"], srna_datas["best"],
                        srna_datas["high"], srna_datas["low"]))
        args_srna.out_table.write("\n")
    args_srna.nums["uni"] += 1


def get_coverage(start, end, strain, wigs, strand, ta, tss, cutoff_coverage,
                 notex, args_srna):
    srna_covers = get_best(wigs, strain, strand, start, end, "total",
                           args_srna, cutoff_coverage)
    srna_datas = replicate_comparison(args_srna, srna_covers, strand, "normal",
                                      None, None, None, notex, cutoff_coverage,
                                      args_srna.texs)
    string = ("\t".join([str(field) for field in [
                             ta.seq_id, "ANNOgesic", "ncRNA", str(start),
                             str(end), ".", ta.strand, "."]]))
    if srna_datas["best"] != 0:
        print_file(string, tss, srna_datas, ta.attributes["sRNA_type"],
                   args_srna, strain)


def check_pro(ta, start, end, srna_datas, type_, cutoff,
              wigs, notex, args_srna):
    '''check the processing site for long non-coding RNA'''
    pro_pos = -1
    detect_pro = "NA"
    for pro in args_srna.pros:
        if (pro.seq_id == ta.seq_id) and (
                pro.strand == ta.strand):
            if ta.strand == "+":
                if (pro.start >= ta.start) and (pro.start <= ta.end) and (
                        (pro.start - start) >= args_srna.min_len) and (
                        (pro.start - start) <= args_srna.max_len):
                    pro_pos = pro.start
                    detect_pro = "".join(["Cleavage:", str(pro.start),
                                          "_", pro.strand])
                if pro.start > ta.end:
                    break
            if ta.strand == "-":
                if (pro.start >= ta.start) and (pro.start <= ta.end) and (
                        (end - pro.start) >= args_srna.min_len) and (
                        (end - pro.start) <= args_srna.max_len):
                    pro_pos = pro.start
                    detect_pro = "".join(["Cleavage:", str(pro.start),
                                          "_", pro.strand])
                    break
                if pro.start > ta.end:
                    break
    new_srna_datas = None
    if ta.strand == "+":
        if ((type_ == "within") and (srna_datas["pos"] < pro_pos)) or (
                (type_ == "longer") and (pro_pos != -1)):
            srna_covers = get_best(wigs, ta.seq_id, ta.strand, start, pro_pos,
                                   "total", args_srna, cutoff)
            new_srna_datas = replicate_comparison(
                    args_srna, srna_covers, ta.strand, "normal", None, None,
                    None, notex, cutoff, args_srna.texs)
            if new_srna_datas["best"] <= cutoff:
                new_srna_datas = None
    else:
        if ((type_ == "within") and (srna_datas["pos"] > pro_pos) and (
                pro_pos != -1)) or (
                (type_ == "longer") and (pro_pos != -1)):
            srna_covers = get_best(wigs, ta.seq_id, ta.strand, pro_pos, end,
                                   "total", args_srna, cutoff)
            new_srna_datas = replicate_comparison(
                    args_srna, srna_covers, ta.strand, "normal", None, None,
                    None, notex, cutoff, args_srna.texs)
            if new_srna_datas["best"] <= cutoff:
                new_srna_datas = None
    return pro_pos, new_srna_datas, detect_pro


def exchange_to_pro(args_srna, srna_datas, ta, start,
                    end, cutoff, wigs, notex):
    detect = False
    if srna_datas["high"] != 0:
        if ((srna_datas["pos"] - start) >= args_srna.min_len) and (
                (srna_datas["pos"] - start) <= args_srna.max_len):
            pro_pos, pro_datas, pro = check_pro(
                    ta, start, end, srna_datas,
                    "within", cutoff, wigs, notex, args_srna)
            if pro_datas is not None:
                srna_datas = pro_datas
                srna_datas["pos"] = pro_pos
                detect = True
            else:
                if srna_datas["best"] > cutoff:
                    detect = True
        else:
            pro_pos, pro_datas, pro = check_pro(
                    ta, start, end, srna_datas,
                    "longer", cutoff, wigs, notex, args_srna)
            if pro_datas is not None:
                srna_datas = pro_datas
                srna_datas["pos"] = pro_pos
                detect = True
    else:
        pro = None
    return detect, srna_datas, pro


def detect_wig_pos(wigs, ta, start, end, tss, cutoff, notex, args_srna):
    '''searching the coverage decrease'''
    srna_covers = get_best(wigs, ta.seq_id, ta.strand, start, end,
                           "differential", args_srna, cutoff)
    srna_datas = replicate_comparison(
            args_srna, srna_covers, ta.strand, "normal", None, None,
            None, notex, cutoff, args_srna.texs)
    detect, srna_datas, pro = exchange_to_pro(args_srna, srna_datas, ta, start,
                                              end, cutoff, wigs, notex)
    if ta.strand == "+":
        if (detect) and (
                (srna_datas["pos"] - start) >= args_srna.min_len) and (
                (srna_datas["pos"] - start) <= args_srna.max_len):
            string = ("\t".join([str(field) for field in [
                      ta.seq_id, "ANNOgesic", "ncRNA", str(start),
                      str(srna_datas["pos"]), ".", ta.strand, "."]]))
            if pro != "NA":
                tss = ";".join([tss, pro])
            print_file(string, tss, srna_datas, ta.attributes["sRNA_type"],
                       args_srna, ta.seq_id)
    else:
        if (detect) and (
                (end - srna_datas["pos"]) >= args_srna.min_len) and (
                (end - srna_datas["pos"]) <= args_srna.max_len):
            string = ("\t".join([str(field) for field in [
                      ta.seq_id, "ANNOgesic", "ncRNA", str(srna_datas["pos"]),
                      str(end), ".", ta.strand, "."]]))
            if pro != "NA":
                tss = ";".join([tss, pro])
            print_file(string, tss, srna_datas, ta.attributes["sRNA_type"],
                       args_srna, ta.seq_id)


def detect_longer(ta, args_srna, cdss, wigs_f, wigs_r):
    '''deal with the long non-coding RNA'''
    notex = None
    if len(args_srna.tsss) != 0:
        for tss in args_srna.tsss:
            cutoff = get_tss_type(tss, args_srna.cutoff_coverage, ta, cdss,
                                  args_srna.file_type, args_srna.break_tran)
            if notex is not None:
                notex = get_tss_type(tss, args_srna.notex, ta, cdss, "notex",
                                     args_srna.break_tran)
            if cutoff is not None:
                if (tss.strand == ta.strand) and (
                        tss.seq_id == ta.seq_id):
                    if (tss.strand == "+"):
                        compare_ta_tss(
                            tss.start, ta.start - args_srna.fuzzy, ta.end, ta,
                            tss, ta.end - tss.start, cutoff, notex,
                            wigs_f, args_srna)
                        if (tss.start >= ta.start - args_srna.fuzzy) and (
                                tss.start <= ta.end) and (
                                (ta.end - tss.start) > args_srna.max_len):
                            if len(wigs_f) != 0:
                                detect_wig_pos(
                                    wigs_f, ta, tss.start, ta.end,
                                    "".join(["TSS:", str(tss.start),
                                             "_", tss.strand]),
                                    cutoff, notex, args_srna)
                    else:
                        compare_ta_tss(
                            tss.end, ta.start, ta.end + args_srna.fuzzy, ta,
                            tss, tss.end - ta.start, cutoff, notex,
                            wigs_r, args_srna)
                        if (tss.end >= ta.start) and (
                                tss.end <= ta.end + args_srna.fuzzy) and (
                                tss.end - ta.start > args_srna.max_len):
                            if len(wigs_r) != 0:
                                detect_wig_pos(
                                    wigs_r, ta, ta.start,
                                    tss.end, "".join(["TSS:", str(tss.end),
                                                      "_", tss.strand]),
                                    cutoff, notex, args_srna)
    if len(args_srna.tsss) == 0:
        cutoff = get_tss_type(None, args_srna.cutoff_coverage, ta, cdss,
                              args_srna.file_type, args_srna.break_tran)
        if (len(wigs_f) != 0) and (len(wigs_r) != 0):
            if ta.strand == "+":
                detect_wig_pos(wigs_f, ta, ta.start, ta.end, "NA",
                               cutoff, notex, args_srna)
            else:
                detect_wig_pos(wigs_r, ta, ta.start, ta.end, "NA",
                               cutoff, notex, args_srna)


def get_tss_type(tss, cutoff_coverage, ta, cdss, file_type, break_tran):
    '''get the cutoff based on the tss type for doing core computation
    If there are multiple TSS type, it will get the high cutoff one'''
    types = []
    for type_, cover in cutoff_coverage.items():
        if cover is not None:
            types.append(type_)
    cover = None
    break_ = False
    if tss is None:
        cover = cutoff_coverage["no_tss"]
    else:
        if ("type" in tss.attributes.keys()):
            for type_ in types:
                if (type_ in tss.attributes["type"].lower()):
                    if cover is None:
                        cover = cutoff_coverage[type_]
                        cover, break_ = check_break_tran(
                                tss, ta, cdss, cover, file_type,
                                break_tran, type_)
                    elif cover < cutoff_coverage[type_]:
                        if (break_) and ((type_.lower() == "primary") or (
                            type_.lower() == "secondary")):
                            pass
                        else:
                            cover = cutoff_coverage[type_]
        else:
            cover = cutoff_coverage["no_tss"]
    return cover


def check_break_tran(tss, ta, cdss, cover, file_type, break_tran, type_):
    '''Check the Primary or Secondary TSS which associated with a 
    non-contain CDS transcript'''
    break_ = False
    if ("type" in tss.attributes.keys()):
        if ("primary" in tss.attributes["type"].lower()) or (
                "secondary" in tss.attributes["type"].lower()):
            overlap = False
            for cds in cdss:
                if (cds.seq_id == ta.seq_id) and (
                        cds.strand == ta.strand):
                    if ((cds.start <= ta.start) and (
                            cds.end >= ta.end)) or (
                            (cds.start >= ta.start) and (
                            cds.end <= ta.end)) or (
                            (cds.start <= ta.start) and (
                            cds.end <= ta.end) and (
                            cds.end >= ta.start)) or (
                            (cds.start >= ta.start) and (
                            cds.start <= ta.end) and (
                            cds.end >= ta.end)):
                        overlap = True
                    elif (cds.start > ta.end):
                        break
            if not overlap:
                break_ = True
                if (file_type == "tex"):
                    tmp_cover = break_tran[0]
                elif (file_type == "notex"):
                    tmp_cover = break_tran[1]
                elif (file_type == "frag"):
                    tmp_cover = break_tran[2]
                if (type_.lower() == "primary") or (
                        type_.lower() == "secondary"):
                    cover = tmp_cover
                else:
                    if tmp_cover > cover:
                        cover = tmp_cover
    return cover, break_


def compare_ta_tss(tss_pos, ta_start, ta_end, ta, tss, diff, cutoff_coverage,
                   notex, wigs, args_srna):
    if (tss_pos >= ta_start) and (tss_pos <= ta_end) and (
            diff >= args_srna.min_len) and (diff <= args_srna.max_len):
        if tss.strand == "+":
            start = tss_pos
            end = ta_end
        else:
            start = ta_start
            end = tss_pos
        if len(wigs) != 0:
            get_coverage(start, end, ta.seq_id, wigs, tss.strand, ta,
                         "".join(["TSS:", str(tss.start), "_", tss.strand]),
                         cutoff_coverage, notex, args_srna)
        else:
            string = "\t".join([str(field) for field in [
                     ta.seq_id, "ANNOgesic", "ncRNA", str(start),
                     str(end), ta.score, ta.strand, ta.phase]])
            print_file(string, "".join(["TSS:", str(tss.start),
                       "_", tss.strand]), None,
                       ta.attributes["sRNA_type"], args_srna, ta.seq_id)
        if args_srna.detects is not None:
            args_srna.detects["uni_with_tss"] = True


def detect_include_tss(ta, args_srna, cdss, wigs_f, wigs_r):
    '''compare sRNA with TSS'''
    args_srna.detects["uni_with_tss"] = False
    notex = None
    for tss in args_srna.tsss:
        cutoff = get_tss_type(tss, args_srna.cutoff_coverage, ta, cdss,
                              args_srna.file_type, args_srna.break_tran)
        if args_srna.notex is not None:
            notex = get_tss_type(tss, args_srna.notex, ta, cdss, "notex",
                                 args_srna.break_tran)
        if cutoff is not None:
            if (tss.strand == ta.strand) and (tss.seq_id == ta.seq_id):
                if (tss.strand == "+"):
                    compare_ta_tss(
                            tss.start, ta.start - args_srna.fuzzy, ta.end,
                            ta, tss, ta.end - tss.start, cutoff, notex,
                            wigs_f, args_srna)
                    if tss.start > ta.end:
                        break
                else:
                    compare_ta_tss(
                            tss.end, ta.start, ta.end + args_srna.fuzzy, ta,
                            tss, tss.end - ta.start, cutoff, notex,
                            wigs_r, args_srna)
                    if (tss.end > ta.end + args_srna.fuzzy):
                        break
    if not args_srna.detects["uni_with_tss"]:
        if (ta.strand == "+") and (len(wigs_f) != 0):
            get_coverage(ta.start, ta.end, ta.seq_id, wigs_f, "+",
                         ta, "False", args_srna.cutoff_coverage["no_tss"],
                         notex, args_srna)
        elif (ta.strand == "-") and (len(wigs_r) != 0):
            get_coverage(ta.start, ta.end, ta.seq_id, wigs_r, "-",
                         ta, "False", args_srna.cutoff_coverage["no_tss"],
                         notex, args_srna)
        elif (len(wigs_f) == 0) and (len(wigs_r) == 0):
            print_file(
                ta.info_without_attributes.replace("Transcript", "ncRNA"),
                "False", None, ta.attributes["sRNA_type"], args_srna,
                ta.seq_id)


def get_proper_tss(tss_file, cutoff_coverage):
    '''get the associated TSS'''
    types = []
    gff_parser = Gff3Parser()
    for type_, cover in cutoff_coverage.items():
        if cover is not None:
            types.append(type_)
    tsss = []
    num_tss = 0
    if tss_file is not None:
        tss_f = open(tss_file, "r")
        for entry in gff_parser.entries(tss_f):
            if ("type" in entry.attributes.keys()):
                for type_ in types:
                    if (type_ in entry.attributes["type"].lower()):
                        tsss.append(entry)
                        num_tss += 1
                        break
            else:
                tsss.append(entry)
                num_tss += 1
        tsss = sorted(tsss, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
        tss_f.close()
    return tsss, num_tss


def read_data(args_srna):
    cdss = []
    tas = []
    pros = []
    genes = []
    ncs = []
    num_cds = 0
    num_ta = 0
    num_pro = 0
    gff_parser = Gff3Parser()
    g_f = open(args_srna.gff_file, "r")
    for entry in gff_parser.entries(g_f):
        if (entry.feature == "CDS") or (
                entry.feature == "pCDS") or (
                entry.feature == "tRNA") or (
                entry.feature == "rRNA"):
            if ("product" in entry.attributes.keys()) and (args_srna.hypo):
                if "hypothetical protein" not in entry.attributes["product"]:
                    cdss.append(entry)
                    num_cds += 1
            else:
                cdss.append(entry)
                num_cds += 1
        if (entry.feature == "gene"):
            genes.append(entry)
        if (entry.feature == "ncRNA"):
            ncs.append(entry)
    if args_srna.pro_file is not None:
        pro_f = open(args_srna.pro_file, "r")
        for entry in gff_parser.entries(pro_f):
            pros.append(entry)
            num_pro += 1
        pros = sorted(pros, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
        pro_f.close()
    t_h = open(args_srna.tran_file)
    for entry_ta in gff_parser.entries(t_h):
        tas.append(entry_ta)
        num_ta += 1
    nums = {"cds": num_cds, "ta": num_ta,
            "pro": num_pro, "uni": 0}
    cdss = sorted(cdss, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    tas = sorted(tas, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    genes = sorted(genes, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    g_f.close()
    t_h.close()
    return nums, cdss, tas, pros, genes, ncs


def read_tss(tss_file):
    tsss = []
    if tss_file is not None:
        tss_f = open(tss_file, "r")
        gff_parser = Gff3Parser()
        for entry in gff_parser.entries(tss_f):
            tsss.append(entry)
        tss_f.close()
    num_tss = None
    return tsss, num_tss


def check_overlap(cds, ta):
    if ((cds.end < ta.end) and (
             cds.end > ta.start) and (
             cds.start <= ta.start)) or (
            (cds.start > ta.start) and (
             cds.start < ta.end) and (
             cds.end >= ta.end)) or (
            (cds.end >= ta.end) and (
             cds.start <= ta.start)) or (
            (cds.end <= ta.end) and (
             cds.start >= ta.start)):
        return True


def compare_ta_cds(cdss, ta, detects):
    for cds in cdss:
        if (cds.strand == ta.strand) and (
                cds.seq_id == ta.seq_id):
            if check_overlap(cds, ta):
                detects["overlap"] = True
                ta.attributes["sRNA_type"] = "in_CDS"
        elif (cds.strand != ta.strand) and (
              cds.seq_id == ta.seq_id):
            if check_overlap(cds, ta):
                detects["anti"] = True
                ta.attributes["sRNA_type"] = "antisense"
    if (not detects["overlap"]) and (not detects["anti"]):
        ta.attributes["sRNA_type"] = "intergenic"


def check_srna_condition(ta, args_srna, cdss, wigs_f, wigs_r):
    '''check the long sRNA and get the coverage of sRNA and check TSS'''
    if ((ta.end - ta.start) >= args_srna.min_len) and (
            (ta.end - ta.start) <= args_srna.max_len):
        if len(args_srna.tsss) != 0:
            detect_include_tss(ta, args_srna, cdss, wigs_f, wigs_r)
        else:
            if (ta.strand == "+") and (len(wigs_f) != 0):
                get_coverage(ta.start, ta.end, ta.seq_id,
                             wigs_f, "+", ta, "NA",
                             args_srna.cutoff_coverage["no_tss"],
                             args_srna.notex["no_tss"], args_srna)
            elif (ta.strand == "-") and (len(wigs_r) != 0):
                get_coverage(ta.start, ta.end, ta.seq_id,
                             wigs_r, "-", ta, "NA",
                             args_srna.cutoff_coverage["no_tss"],
                             args_srna.notex["no_tss"], args_srna)
            if (len(wigs_f) == 0) and (len(wigs_r) == 0):
                print_file(ta.info_without_attributes.replace("Transcript",
                           "ncRNA"), "NA", None, ta.attributes["sRNA_type"],
                           args_srna, ta.seq_id)
    if ((ta.end - ta.start) > args_srna.max_len):
        detect_longer(ta, args_srna, cdss, wigs_f, wigs_r)


def get_cutoff(cutoffs, out_folder, file_type):
    '''set the cutoff of intergenic and antisense sRNA'''
    out = open(os.path.join(out_folder, "tmp_cutoff_inter"), "a")
    coverages = {}
    num_cutoff = 0
    for cutoff in cutoffs:
        if (cutoff != "0") and (num_cutoff == 0):
            coverages["primary"] = float(cutoff)
        elif (cutoff != "0") and (num_cutoff == 1):
            coverages["secondary"] = float(cutoff)
        elif (cutoff != "0") and (num_cutoff == 2):
            coverages["internal"] = float(cutoff)
        elif (cutoff != "0") and (num_cutoff == 3):
            coverages["antisense"] = float(cutoff)
        elif (cutoff != "0") and (num_cutoff == 4):
            coverages["orphan"] = float(cutoff)
        num_cutoff += 1
    low = None
    for cover in coverages.values():
        if cover != 0:
            if low is None:
                low = cover
            elif cover < float(low):
                low = cover
    if low is None:
        print("Error: The cutoff of coverage can not be all 0...")
        sys.exit()
    coverages["no_tss"] = float(low)
    for tss, cover in coverages.items():
        out.write("\t".join([file_type, tss, str(cover)]) + "\n")
    out.close()
    return coverages


def compute_tss_type(args_srna, cdss, genes, wigs_f, wigs_r):
    tsss, num_tss = read_tss(args_srna.tss_file)
    if "TSS_classes" not in os.listdir(args_srna.out_folder):
        os.mkdir(os.path.join(args_srna.out_folder, "TSS_classes"))
    new_tss_file = os.path.join(args_srna.out_folder, "TSS_classse",
                                "_".join([args_srna.prefix, "TSS.gff"]))
    new_tss_fh = open(new_tss_file, "w")
    num_tss = 0
    for tss in tsss:
        tss_type = compare_tss_cds(tss, cdss, genes)
        tss.attributes = tss_type[1]
        tss.attributes["ID"] = tss.seq_id + "_tss" + str(num_tss)
        tss.attribute_string = "".join([tss_type[0],
            ";ID=", tss.seq_id, "_tss", str(num_tss)])
        num_tss += 1
    final_tsss = fix_primary_type(tsss, wigs_f, wigs_r)
    for tss in final_tsss:
        tss.attribute_string = ";".join(
            ["=".join(items) for items in tss.attributes.items()])
        new_tss_fh.write("\t".join([str(field) for field in [
                         tss.seq_id, tss.source, tss.feature, tss.start,
                         tss.end, tss.score, tss.strand, tss.phase,
                         tss.attribute_string]]) + "\n")
    new_tss_fh.close()


def get_intergenic_antisense_cutoff(args_srna):
    '''set the cutoff of intergenic and antisense sRNA
    also deal with the no tex library'''
    cutoff_coverage = get_cutoff(args_srna.cutoffs, args_srna.out_folder,
                                 args_srna.file_type)
    notex = None
    if args_srna.cut_notex is not None:
        notex = get_cutoff(args_srna.cut_notex, args_srna.out_folder, "notex")
    return cutoff_coverage, notex


def free_memory(paras):
    for data in paras:
        del(data)
    gc.collect()


def check_ncRNA(gene, ncs):
    detect = False
    for nc in ncs:
        if "Parent" in nc.attributes.keys():
            if gene.attributes["ID"] in nc.attributes["Parent"]:
                detect = True
        elif (nc.seq_id == gene.seq_id) and (
                nc.strand == gene.strand):
            if (gene.start <= nc.start) and (
                    gene.end >= nc.end):
                detect = True
    return detect


def intergenic_srna(args_srna, libs, texs, wigs_f, wigs_r):
    '''get intergenic and antisense sRNA'''
    inter_cutoff_coverage, inter_notex = get_intergenic_antisense_cutoff(
                                         args_srna)
    anti_cutoff_coverage, anti_notex = get_intergenic_antisense_cutoff(
                                       args_srna)
    nums, cdss, tas, pros, genes, ncs = read_data(args_srna)
    if not args_srna.tss_source:
        print("Classifying TSSs...")
        compute_tss_type(args_srna, cdss, genes, wigs_f, wigs_r)
        print("Classification of TSSs has done...")
    tsss, num_tss = read_tss(args_srna.tss_file)
    detects = {"overlap": False, "uni_with_tss": False, "anti": False}
    output = open(args_srna.output_file, "w")
    out_table = open(args_srna.output_table, "w")
    output.write("##gff-version 3\n")
    for ta in tas:
        detects["overlap"] = False
        detects["anti"] = False
        compare_ta_cds(cdss, ta, detects)
        if (detects["overlap"]) and (not args_srna.in_cds):
            continue
        else:
            if not detects["anti"]:
                cutoff_coverage = inter_cutoff_coverage
                notex = inter_notex
            else:
                cutoff_coverage = anti_cutoff_coverage
                notex = anti_notex
            args_srna = ArgsContainer().extend_inter_container(
                            args_srna, tsss, pros, nums,
                            output, out_table, texs, detects,
                            cutoff_coverage, notex)
            check_srna_condition(ta, args_srna, cdss, wigs_f, wigs_r)
    file_name = args_srna.output_file.split(".")
    file_name = file_name[0] + ".stat"
    output.close()
    out_table.close()
    paras = [tsss, tas, pros, genes, cdss]
    free_memory(paras)



import csv
import numpy as np
import os
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.coverage_detection import coverage_comparison, check_tex, get_repmatch
from annogesiclib.lib_reader import read_libs, read_wig


def import_data(row):
    return {"method": "gene_converged", "strain": row[0],
            "start": int(row[1]), "end": int(row[2]), "name": row[3],
            "miss": int(row[4]), "loop": int(row[5]), "diff": [],
            "length": int(row[6]), "r_stem": int(row[7]), "strand": row[8],
            "l_stem": int(row[9]), "parent_p": row[10], "parent_m": row[11],
            "ut": int(row[12]), "print": False, "detect_p": False,
            "detect_m": False, "express": "False"}


def compare_ta(terms, tas, fuzzy):
    '''Compare transcript and terminator to 
    find the expressed terminator'''
    for term in terms:
        for ta in tas:
            start = ta.start - fuzzy
            end = ta.end + fuzzy
            if ((ta.seq_id == term["strain"]) and (
                    ta.strand == term["strand"])):
                if ((start <= term["start"]) and (
                         end >= term["start"]) and (
                         end <= term["end"])) or (
                        (start >= term["start"]) and (
                         end <= term["end"])) or (
                        (start >= term["start"]) and (
                         start <= term["end"]) and (
                         end >= term["end"])) or (
                        (start <= term["start"]) and (
                         end >= term["end"])):
                    term["express"] = "True"


def compare_transtermhp(hps, fr_terms):
    '''compare two methods - gene converged and TransTermHP'''
    terms = []
    for term in fr_terms:
        detect = False
        for hp in hps:
            if (hp.seq_id == term["strain"]) and (
                    hp.strand == term["strand"]):
                if ((hp.start <= term["start"]) and (
                         hp.end >= term["start"]) and (
                         hp.end <= term["end"])) or (
                        (hp.start >= term["start"]) and (
                         hp.end <= term["end"])) or (
                        (hp.start >= term["start"]) and (
                         hp.start <= term["end"]) and (
                         hp.end >= term["end"])) or (
                        (hp.start <= term["start"]) and (
                         hp.end >= term["end"])):
                    if hp.start < term["start"]:
                        term["start"] = hp.start
                    if hp.end > term["end"]:
                        term["end"] = hp.end
                    hp.attributes["print"] = True
                    detect = True
        if detect:
            term["method"] = ",".join([term["method"], "TransTermHP"])
        terms.append(term)
    for hp in hps:
        need_print = False
        if "print" not in hp.attributes.keys():
            need_print = True
        else:
            if not hp.attributes["print"]:
                need_print = True
        if need_print:
            if hp.strand == "+":
                terms.append({"method": "TransTermHP", "strain": hp.seq_id,
                              "start": hp.start, "end": hp.end,
                              "strand": hp.strand, "name": hp.attributes["ID"],
                              "parent_p": hp.attributes["associated_gene"],
                              "print": False, "detect_p": False,
                              "detect_m": False, "express": "False",
                              "diff": []})
            else:
                terms.append({"method": "TransTermHP", "strain": hp.seq_id,
                              "start": hp.start, "end": hp.end,
                              "strand": hp.strand, "name": hp.attributes["ID"],
                              "parent_m": hp.attributes["associated_gene"],
                              "print": False, "detect_p": False,
                              "detect_m": False, "express": "False",
                              "diff": []})
    terms = sorted(terms, key=lambda x: (x["strain"], x["start"],
                                         x["end"], x["strand"]))
    return terms

def compare_replicates(term_covers, template_texs, cond, args_term):
    '''check the cutoff of replicate match'''
    detect_num = 0
    term_datas = []
    diff_cover = -1
    diff = []
    detect = False
    detect_num = check_tex(template_texs, term_covers, term_datas, None,
                           "terminator", None, None, None, None, 0,
                           args_term.tex_notex)

    if ("texnotex" in cond):
        tex_rep = get_repmatch(args_term.replicates["tex"], cond)
        if detect_num >= tex_rep:
            detect = True
    elif ("frag" in cond):
        frag_rep = get_repmatch(args_term.replicates["frag"], cond)
        if detect_num >= frag_rep:
            detect = True
    if detect:
        detect = False
        for term in term_datas:
            if (len(diff) == 0) or (diff_cover < term["diff"]):
                diff_cover = term["diff"]
                diff = term
    return diff_cover, diff, term_datas, detect_num


def coverage2term(covers, term, hl_covers, hl_poss, strand, term_covers,
                  track, args_term, start_plus, end_minus, lib_type):
    '''It is for get the highest and lowest coverage'''
    first = True
    pos = 0
    for cover in covers:
        if strand == "+":
            cover_pos = start_plus + pos
        else:
            cover_pos = end_minus - pos
        if (term["start"] <= cover_pos + args_term.fuzzy) and (
                term["end"] >= cover_pos - args_term.fuzzy):
            first = coverage_comparison(cover, hl_covers, hl_poss,
                                        first, strand, cover_pos)
        else:
            if (strand == "+") and (
                    cover_pos > term["end"] + args_term.fuzzy):
                break
            elif (strand == "-") and (
                    cover_pos < term["start"] - args_term.fuzzy):
                break
        if (first is not True) and (hl_covers["high"] > 0):
            if ((hl_covers["low"] / hl_covers["high"]) <
                    args_term.decrease) and (
                    hl_covers["low"] > -1):
                term_covers.append({
                    "track": track, "high": hl_covers["high"],
                    "low": hl_covers["low"], "detect": "True",
                    "diff": (hl_covers["high"] - hl_covers["low"]),
                    "type": lib_type})
                break
        pos += 1


def check_start_and_end(term, args_term, covers):
    if (term["start"] - args_term.fuzzy - 2) < 0:
        start = 0
    else:
        start = term["start"] - args_term.fuzzy - 2
    if (term["end"] + args_term.fuzzy + 1) > len(covers):
        end = len(covers)
    else:
        end = term["end"] + args_term.fuzzy + 1
    return start, end

def get_coverage(term, wigs, strand, template_texs, args_term):
    '''get proper coverage to check the coverage decrease'''
    hl_poss = {"high": 0, "low": 0}
    hl_covers = {"high": 0, "low": 0}
    term_datas = {}
    detect_nums = {}
    diff_cover = -1
    diff = []
    for wig_strain, conds in wigs.items():
        if wig_strain == term["strain"]:
            for cond, tracks in conds.items():
                term_covers = []
                term_datas[cond] = []
                detect_nums[cond] = 0
                for lib_name, covers in tracks.items():
                    track = lib_name.split("|")[-3]
                    lib_strand = lib_name.split("|")[-2]
                    lib_type = lib_name.split("|")[-1]
                    c_start, c_end = check_start_and_end(term, args_term, covers)
                    covers = covers[c_start: c_end]
                    if strand == "-":
                        covers = covers[::-1]
                    coverage2term(covers, term, hl_covers, hl_poss,
                                  strand, term_covers, track, args_term,
                                  c_start, c_end, lib_type)
                if len(term_covers) != 0:
                    tmp_cov, tmp_diff, term_datas[cond], detect_nums[cond] = (
                            compare_replicates(term_covers, template_texs,
                                               cond, args_term))
                    if (diff_cover == -1) or (diff_cover < tmp_cov):
                        diff_cover = tmp_cov
                        diff = tmp_diff
    detect = False
    for cond, num in detect_nums.items():
        if ("texnotex" in cond):
            tex_rep = get_repmatch(args_term.replicates["tex"], cond)
            if num >= tex_rep:
                detect = True
        elif ("frag" in cond):
            frag_rep = get_repmatch(args_term.replicates["frag"], cond)
            if num >= frag_rep:
                detect = True
        if detect:
            detect = False
            if strand == "+":
                term["detect_p"] = True
            else:
                term["detect_m"] = True
    return diff_cover, diff, term_datas, detect_nums


def compare_term(term, terms):
    '''For the terminators which are associated with the gene, it 
    will try to find the best one based on secondary structure'''
    if len(terms) != 0:
        for tmp in terms:
            if term["miss"] < tmp["miss"]:
                terms = []
                terms.append(term)
                break
            elif term["miss"] == tmp["miss"]:
                if ("diff_cover" in term.keys()) and \
                   ("diff_cover" in tmp.keys()):
                    if (term["diff_cover"] > tmp["diff_cover"]):
                        terms = []
                        terms.append(term)
                    elif (term["diff_cover"] == tmp["diff_cover"]):
                        if term["ut"] > tmp["ut"]:
                            terms = []
                            terms.append(term)
                        elif term["ut"] == tmp["ut"]:
                            terms.append(term)
                break
            elif term["miss"] > tmp["miss"]:
                break
    else:
        terms.append(term)
    return terms


def first_term(strand, term, detect_terms, detect):
    '''for the first and latest terminators, 
    we only need to check parent gene for one side of terminator
    and also need to take care of the terminal end'''
    if (strand == "+"):
        if (term["detect_p"]):
            detect_terms["detect"].append(term)
            detect = True
        else:
            detect_terms["undetect"].append(term)
    elif (strand == "-"):
        if (term["detect_m"]):
            detect_terms["detect"].append(term)
            detect = True
        else:
            detect_terms["undetect"].append(term)
    return detect


def get_attribute_string(num, name, parent, diff, term, coverage, method):
    attribute_string = ";".join(
                 ["=".join(items) for items in [
                  ("ID", term["strain"] + "_terminator" + str(num)),
                  ("Name", name), ("associated_gene", parent),
                  ("coverage_decrease", coverage),
                  ("diff_coverage", diff),
                  ("express", term["express"]),
                  ("method", method)]])
    return attribute_string


def print_table(term, out_t, args_term):
    '''Print to table'''
    first = True
    if (term["express"] == "True") and \
       (term["diff_cover"] != -1):
        if term["diff"]["high"] >= args_term.cutoff_coverage:
            out_t.write("\tTrue\t")
            if not args_term.table_best:
                for datas in term["datas"].values():
                    for data in datas:
                        if first:
                            out_t.write(
                                "{0}(diff={1};high={2};low={3})".format(
                                    data["track"], data["diff"],
                                    data["high"], data["low"]))
                            first = False
                        else:
                            out_t.write(
                                ";{0}(diff={1};high={2};low={3})".format(
                                    data["track"], data["diff"],
                                    data["high"], data["low"]))
            else:
                out_t.write("{0}(diff={1};high={2};low={3})".format(
                            term["diff"]["track"], term["diff_cover"],
                            term["diff"]["high"], term["diff"]["low"]))
        else:
            out_t.write("\tFalse\t")
            out_t.write("No_coverage_decreasing")
    elif (term["express"] == "True") and \
         (term["diff_cover"] == -1):
        out_t.write("\tFalse\t")
        out_t.write("No_coverage_decreasing")
    elif term["express"] == "False":
        out_t.write("\tFalse\t")
        out_t.write("NA")


def print2file(num, term, coverage, parent, out, out_t, method, args_term):
    '''Print to gff file and table'''
    name = 'terminator_%0*d' % (5, num)
    if ("detect_num" in term.keys()) and \
       (term["diff_cover"] != -1):
        out_t.write("\t".join([term["strain"], name, str(term["start"]),
                              str(term["end"]), term["strand"],
                              term["method"].replace(",", ";"), parent]))
    else:
        out_t.write("\t".join([term["strain"], name, str(term["start"]),
                              str(term["end"]), term["strand"],
                              term["method"].replace(",", ";"), parent]))
    if (term["express"] == "True") and (term["diff_cover"] != -1):
        if (term["diff"]["high"] >= args_term.cutoff_coverage):
            diff = ("{0}(high:{1},low:{2})".format(
                    term["diff"]["track"], term["diff"]["high"],
                    term["diff"]["low"]))
            attribute_string = get_attribute_string(
                num, name, parent, diff, term, coverage, method)
        elif (term["diff"]["high"] < args_term.cutoff_coverage):
            attribute_string = get_attribute_string(
                num, name, parent, "NA", term,
                "No_coverage_decreasing", method)
    elif (term["express"] == "True") and (term["diff_cover"] == -1):
        attribute_string = get_attribute_string(
            num, name, parent, "NA", term, "No_coverage_decreasing", method)
    elif (term["express"] == "False"):
        attribute_string = get_attribute_string(
            num, name, parent, "NA", term, "NA", method)
    out.write("\t".join([str(field) for field in [
              term["strain"], "ANNOgesic", "terminator",
              str(term["start"]), str(term["end"]), ".",
              term["strand"], ".", attribute_string]]) + "\n")
    print_table(term, out_t, args_term)
    out_t.write("\n")


def print_detect_undetect(term, num, out, out_t, detect, args_term):
    '''For controlling the output of different strand of terminator'''
    if term["strand"] == "+":
        print2file(num, term, detect, term["parent_p"], out,
                   out_t, term["method"], args_term)
        num += 1
    else:
        print2file(num, term, detect, term["parent_m"], out,
                   out_t, term["method"], args_term)
        num += 1
    return num


def term_validation(pre_term, term, detect, detect_terms, out,
                    out_t, num, args_term, final_terms):
    '''Classification of terminators'''
    if (pre_term["name"] != term["name"]) or (args_term.keep_multi):
        if detect:
            final_terms["detect"] = (final_terms["detect"] +
                                     detect_terms["detect"])
            detect = False
        else:
            final_terms["undetect"] = (final_terms["undetect"] +
                                       detect_terms["undetect"])
        detect_terms["detect"] = []
        detect_terms["undetect"] = []
        detect = first_term(term["strand"], term, detect_terms, detect)
    else:
        if term["strand"] == "+":
            if (term["detect_p"]):
                detect = True
                detect_terms["detect"] = compare_term(
                                         term, detect_terms["detect"])
            else:
                if not detect:
                    detect_terms["undetect"] = compare_term(
                                               term, detect_terms["undetect"])
        else:
            if (term["detect_m"]):
                detect = True
                detect_terms["detect"] = compare_term(
                                         term, detect_terms["detect"])
            else:
                if not detect:
                    detect_terms["undetect"] = compare_term(
                                               term, detect_terms["undetect"])
    return num, detect


def print_term(terms, out, out_t, args_term):
    first = True
    detect = False
    detect_terms = {"detect": [], "undetect": []}
    num = 0
    final_terms = {"detect": [], "undetect": []}
    for term in terms:
        if first:
            first = False
            pre_term = term
            detect = first_term(term["strand"], term, detect_terms, detect)
        else:
            num, detect = term_validation(pre_term, term, detect, detect_terms,
                                          out, out_t, num, args_term,
                                          final_terms)
            pre_term = term
    if detect:
        final_terms["detect"] = final_terms["detect"] + detect_terms["detect"]
    else:
        final_terms["undetect"] = (final_terms["undetect"] + 
                                   detect_terms["undetect"])
    remove_repeat(final_terms["detect"], num, out, out_t,
                  "True", args_term)
    remove_repeat(final_terms["undetect"], num, out, out_t,
                  "False", args_term)


def del_repeat_term(terms):
    '''delete the repeat terminators'''
    first = True
    new_terms = []
    for term in terms:
        detect = False
        if first:
            first = False
            pre_term = term
        else:
            if (term["strain"] == pre_term["strain"]) and (
                    term["strand"] == pre_term["strand"]) and (
                    term["parent_p"] == pre_term["parent_p"]) and (
                    term["parent_m"] == pre_term["parent_m"]):
                if (term["start"] <= pre_term["start"]) and (
                        term["end"] >= pre_term["start"]) and (
                        term["end"] <= pre_term["end"]):
                    detect = True
                    pre_term["start"] = term["start"]
                elif (term["start"] <= pre_term["start"]) and (
                        term["end"] >= pre_term["end"]):
                    detect = True
                    pre_term["start"] = term["start"]
                    pre_term["end"] = term["end"]
                elif (term["start"] >= pre_term["start"]) and (
                        term["end"] <= pre_term["end"]):
                    detect = True
                elif (term["start"] >= pre_term["start"]) and (
                        term["start"] <= pre_term["end"]) and (
                        term["end"] >= pre_term["end"]):
                    detect = True
                    pre_term["end"] = term["end"]
                if detect:
                    if term["miss"] < pre_term["miss"]:
                        pre_term["miss"] = term["miss"]
                else:
                    new_terms.append(pre_term)
                    pre_term = term
            else:
                new_terms.append(pre_term)
                pre_term = term
    new_terms.append(term)
    return new_terms


def read_data(gff_file, tran_file, tranterm_file, seq_file, term_table):
    gff_parser = Gff3Parser()
    gffs = []
    tas = []
    hps = []
    fr_terms = []
    seq = {}
    new_terms = []
    for entry in gff_parser.entries(open(gff_file)):
        if (entry.feature == "gene"):
            gffs.append(entry)
    if os.path.exists(tran_file):
        for entry in gff_parser.entries(open(tran_file)):
            tas.append(entry)
    if os.path.exists(tranterm_file):
        for entry in gff_parser.entries(open(tranterm_file)):
            hps.append(entry)
    with open(seq_file, "r") as s_f:
        for line in s_f:
            line = line.strip()
            if line.startswith(">"):
                strain = line[1:]
                seq[strain] = ""
            else:
                seq[strain] = seq[strain] + line
    if os.path.exists(term_table):
        term_f = open(term_table, "r")
        for row in csv.reader(term_f, delimiter="\t"):
            fr_terms.append(import_data(row))
        new_terms = del_repeat_term(fr_terms)
    tas = sorted(tas, key=lambda x: (x.seq_id, x.start, x.end, x.strand))
    gffs = sorted(gffs, key=lambda x: (x.seq_id, x.start, x.end, x.strand))
    hps = sorted(hps, key=lambda x: (x.seq_id, x.start, x.end, x.strand))
    return gffs, tas, hps, new_terms, seq


def compute_wig(wig_file, libs, terms, strand, texs, args_term):
    '''Get the coverage information to specific terminator'''
    wigs = {}
    wigs = read_wig(wig_file, strand, libs)
    for term in terms:
        if (term["strand"] == strand) and \
           (term["express"] == "True"):
            term["diff_cover"], term["diff"], term["datas"], detect_nums = \
                get_coverage(term, wigs, strand, texs, args_term)
            term["detect_num"] = {}
            for cond, num in detect_nums.items():
                term["detect_num"][cond] = str(num)



def remove_repeat(terms, num, out, out_t, detect, args_term):
    finals = []
    for term1 in terms:
        if args_term.keep_multi:
            num = print_detect_undetect(term1, num, out, out_t,
                                        detect, args_term)
        else:
            tmp_term = term1
            for term2 in terms:
                if (term1["strain"] == term2["strain"]) and (
                        term1["strand"] == term2["strand"]):
                    if term1["strand"] == "+":
                        parents1 = set(tmp_term["parent_p"].split(","))
                        parents2 = set(term2["parent_p"].split(","))
                        if (parents1.issubset(parents2)):
                            if (parents2.issubset(parents1)):
                                if tmp_term["end"] < term2["end"]:
                                    tmp_term = term2
                                elif (tmp_term["end"] == term2["end"]) and (
                                        tmp_term["start"] > term2["start"]):
                                    tmp_term = term2
                            else:
                                tmp_term = term2
                    else:
                        parents1 = set(tmp_term["parent_m"].split(","))
                        parents2 = set(term2["parent_m"].split(","))
                        if (parents1.issubset(parents2)):
                            if (parents2.issubset(parents1)):
                                if tmp_term["start"] > term2["start"]:
                                    tmp_term = term2
                                elif (tmp_term["start"] == term2["start"]) and (
                                        term1["end"] < term2["end"]):
                                    tmp_term = term2
                            else:
                                tmp_term = term2
            if tmp_term not in finals:
                num = print_detect_undetect(tmp_term, num, out, out_t,
                                            detect, args_term)
                finals.append(tmp_term)
    return num


def detect_coverage(term_table, gff_file, tran_file, seq_file,
                    wig_f_file, wig_r_file, tranterm_file, wig_folder,
                    output_file, output_table, args_term):
    '''For detecting the coverage of terminator'''
    gffs, tas, hps, fr_terms, seq = read_data(gff_file, tran_file,
                                              tranterm_file, seq_file,
                                              term_table)
    terms = compare_transtermhp(hps, fr_terms)
    compare_ta(terms, tas, args_term.fuzzy)
    libs, texs = read_libs(args_term.libs, wig_folder)
    compute_wig(wig_f_file, libs, terms, "+", texs, args_term)
    compute_wig(wig_r_file, libs, terms, "-", texs, args_term)
    out = open(output_file, "w")
    out_t = open(output_table, "w")
    print_term(terms, out, out_t, args_term)



from annogesiclib.gff3 import Gff3Parser
from annogesiclib.helper import Helper


def get_feature(cds):
    if "locus_tag" in cds.attributes.keys():
        feature = cds.attributes["locus_tag"]
    elif "protein_id" in cds.attributes.keys():
        feature = cds.attributes["protein_id"]
    else:
        feature = cds.attributes["ID"]
    return feature


def import_data(seq, cds, start, end):
    feature = get_feature(cds)
    return {"seq": seq, "strain": cds.seq_id, "strand": cds.strand,
            "protein": feature, "start": start, "end": end}


def detect_site(inters, args_ribo):
    '''Detection of ribosome binding site'''
    rbss = []
    for inter in inters:
        for nts in range(0, len(inter["seq"]) - 6):
            num = 0
            miss = 0
            detect = False
            for nt in inter["seq"][nts:nts + 6]:
                if miss > args_ribo.fuzzy_rbs:
                    break
                else:
                    if (num == 0) and (nt != "A"):
                        miss += 1
                    elif (num == 1) and (nt != "G"):
                        miss += 1
                    elif (num == 2) and (nt != "G"):
                        miss += 1
                    elif (num == 3) and (nt != "A"):
                        miss += 1
                    elif (num == 4) and (nt != "G"):
                        miss += 1
                    elif (num == 5) and (nt != "G"):
                        miss += 1
                    num += 1
            if miss <= args_ribo.fuzzy_rbs:
                for start_codon in args_ribo.start_codons:
                    if (start_codon in inter["seq"][
                            nts + (6 + args_ribo.start_rbs - 1):
                            nts + (6 + args_ribo.end_rbs)]):
                        rbss.append(inter)
                        detect = True
                        break
            if detect:
                break
    return rbss


def read_file(seq_file, gff_file, tss_file, tran_file):
    cdss = []
    tsss = []
    trans = []
    seq = {}
    with open(seq_file, "r") as f_h:
        for line in f_h:
            line = line.strip()
            if line.startswith(">"):
                strain = line[1:]
                seq[strain] = ""
            else:
                seq[strain] = seq[strain] + line
    g_h = open(gff_file)
    for entry in Gff3Parser().entries(g_h):
        if (entry.feature == "CDS"):
            cdss.append(entry)
    t_h = open(tss_file)
    for entry in Gff3Parser().entries(t_h):
        tsss.append(entry)
    a_h = open(tran_file)
    for entry in Gff3Parser().entries(a_h):
        trans.append(entry)
    cdss = sorted(cdss, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    tsss = sorted(tsss, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    trans = sorted(trans, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    g_h.close()
    a_h.close()
    t_h.close()
    return cdss, seq, tsss, trans


def extract_inter_seq(inter, cds, seq, fuzzy, inters):
    '''extract the sequence of pre-CDS region'''
    helper = Helper()
    start = inter["start"] - fuzzy
    end = inter["end"] + fuzzy
    if inter["start"] - fuzzy <= 0:
        start = 1
    if inter["end"] + fuzzy >= len(seq[cds.seq_id]):
        end = len(seq)
    if cds.strand == "+":
        inter_seq = helper.extract_gene(seq[cds.seq_id], start,
                                        end, "+")
    else:
        inter_seq = helper.extract_gene(seq[cds.seq_id], start,
                                        end, "-")
    inters.append(import_data(inter_seq, cds, inter["start"], inter["end"]))


def compare_tss(tsss, cds, inters, fuzzy, seq, utr):
    '''Compare CDS with TSS to get the pre-CDS region'''
    for tss in tsss:
        if (cds.seq_id == tss.seq_id) and (
                cds.strand == tss.strand):
            if tss.strand == "+":
                if ((cds.start - tss.start) <= utr) and (
                        (cds.start - tss.start) > 0):
                    inter = {"start": tss.start, "end": cds.start,
                             "strain": cds.seq_id, "strand": cds.strand}
                    extract_inter_seq(inter, cds, seq, fuzzy, inters)
            else:
                if ((tss.end - cds.end) <= utr) and (
                        (tss.end - cds.end) > 0):
                    inter = {"start": cds.end, "end": tss.end,
                             "strain": cds.seq_id, "strand": cds.strand}
                    extract_inter_seq(inter, cds, seq, fuzzy, inters)


def compare_pre_cds(first, cdss, cds, seq):
    '''Search the front position CDS of the query one 
    to get the pre-CDS region'''
    detect_cds = False
    for pre_cds in cdss:
        if (pre_cds.seq_id == cds.seq_id) and (
                pre_cds.strand == cds.strand):
            if pre_cds.strand == "+":
                if first:
                    start = 1
                    end = cds.start
                    detect_cds = True
                    first = False
                    break
                elif pre_cds.end < cds.start:
                    start = pre_cds.end
                    end = cds.start
                    detect_cds = True
                elif pre_cds.end >= cds.start:
                    break
            else:
                if pre_cds.start > cds.end:
                    start = cds.end
                    end = pre_cds.start
                    detect_cds = True
                    break
    if (not detect_cds) and (cds.strand == "-"):
        start = cds.end
        end = len(seq[cds.seq_id])
    return first, start, end


def compare_tran(cds, trans, seq, inters, fuzzy, start, end):
    '''For detect the expressed region of candidates'''
    detect = False
    for tran in trans:
        if (tran.seq_id == cds.seq_id) and (
                tran.strand == cds.strand):
            if tran.strand == "+":
                if (cds.start > tran.start) and (
                        cds.start <= tran.end):
                    if start < tran.start:
                        start = tran.start
                        detect = True
                elif (tran.start > cds.start):
                    break
            else:
                if (cds.end > tran.start) and (
                        cds.end <= tran.end):
                    if end > tran.start:
                        end = tran.start
                        detect = True
                elif (tran.start > cds.end):
                    break
    if detect:
        inter = {"start": tran.start, "end": cds.start,
                 "strain": cds.seq_id, "strand": cds.strand}
        extract_inter_seq(inter, cds, seq, fuzzy, inters)
    else:
        for tran in trans:
            if (tran.seq_id == cds.seq_id) and (
                    tran.strand == cds.strand):
                if ((start <= tran.start) and (
                         end >= tran.end)) or (
                        (start >= tran.start) and (
                         end <= tran.end)) or (
                        (start <= tran.start) and (
                         end <= tran.end) and (
                         end >= tran.start)) or (
                        (start >= tran.start) and (
                         start <= tran.end) and (
                         end >= tran.end)):
                    inter = {"start": tran.start, "end": cds.start,
                             "strain": cds.seq_id, "strand": cds.strand}
                    extract_inter_seq(inter, cds, seq, fuzzy, inters)
                    break


def extract_seq(cdss, seq, tsss, trans, fuzzy, utr):
    '''extract the sequence for searching the riboswitch or RNA thermometer
    by comparing with TSS, transcript and CDS'''
    first = True
    inters = []
    for cds in cdss:
        compare_tss(tsss, cds, inters, fuzzy, seq, utr)
        first, start, end = compare_pre_cds(first, cdss, cds, seq)
        compare_tran(cds, trans, seq, inters, fuzzy, start, end)
    return inters


def extract_potential_rbs(seq_file, gff_file, tss_file, tran_file,
                          out_file, args_ribo, feature):
    '''Get the potential riboswitch or RNA-thermometer'''
    out = open(out_file, "w")
    cdss, seq, tsss, trans = read_file(seq_file, gff_file, tss_file, tran_file)
    inters = extract_seq(cdss, seq, tsss, trans,
                         args_ribo.fuzzy, args_ribo.utr)
    rbss = detect_site(inters, args_ribo)
    num = 0
    for rbs in rbss:
        out.write(">" + feature + "_{0}\n".format(
                  "|".join([str(num), rbs["strain"], rbs["strand"],
                            rbs["protein"], str(rbs["start"]),
                            str(rbs["end"])])))
        out.write(rbs["seq"] + "\n")
        num += 1
    out.close()



import sys
from annogesiclib.gff3 import Gff3Parser


def import_psortb(seq_name, psortbs, local_name, local_score, type_, results):
    seq_datas = seq_name.split("__")
    seq_id = seq_datas[0]
    features = seq_datas[1].split("_")
    prot_id = "_".join(features[:-3])
    if type_ == "multi":
        psortbs.append({"seq_id": seq_id,
                        "protein_id": prot_id,
                        "strand": features[-3],
                        "start": int(features[-2]),
                        "end": int(features[-1]),
                        "local": "/".join(local_name),
                        "score": "/".join(local_score)})
    else:
        psortbs.append({"seq_id": seq_id,
                        "protein_id": prot_id,
                        "strand": features[-3],
                        "start": int(features[-2]),
                        "end": int(features[-1]),
                        "local": results[0],
                        "score": results[-1]})
    return {"datas": seq_datas, "features": features, "prot_id": prot_id}


def get_results(line, scores, psortbs, out_p, seq_name, fuzzy):
    '''print the results of psorb'''
    local_name = []
    local_score = []
    if len(line) == 0:
        pass
    elif "(This protein may have multiple localization sites.)" in line:
        results = line.split(" ")
        sort_scores = sorted(scores, key=lambda x: (x["score"]), reverse=True)
        first = True
        high_scores = []
        for score in sort_scores:
            if first:
                high_scores.append(score)
                first = False
                best_score = score
            else:
                if score["local"] != results[0]:
                    if score["score"] < (best_score["score"] - fuzzy):
                        break
                    else:
                        high_scores.append(score)
        for high_score in high_scores:
            local_name.append(high_score["local"])
            local_score.append(str(high_score["score"]))
        seq_datas = import_psortb(seq_name, psortbs, local_name,
                                  local_score, "multi", results)
        out_p.write("\t".join([seq_datas["datas"][0], seq_datas["prot_id"],
                               "\t".join(seq_datas["features"][-3:]),
                               "/".join(local_name),
                               "/".join(local_score)]) + "\n")
    else:
        results = line.split(" ")
        seq_datas = import_psortb(seq_name, psortbs, None,
                                  None, "unique", results)
        out_p.write("\t".join([seq_datas["datas"][0], seq_datas["prot_id"],
                               "\t".join(seq_datas["features"][-3:]),
                               results[0], results[-1]]) + "\n")
    return local_name, local_score


def get_information(psortb_table, out_p, fuzzy):
    '''get the information of psorb'''
    scores = []
    psortbs = []
    seq_name = None
    detects = {"score": False, "result": False}
    with open(psortb_table, "r") as p_h:
        for line in p_h:
            line = line.strip()
            if (line.startswith("--")) or \
               (line.startswith("Secondary localization(s):")):
                detects["result"] = False
            if detects["score"]:
                if "Final Prediction:" not in line:
                    datas = line.split(" ")
                    scores.append({"local": datas[0],
                                   "score": float(datas[-1])})
            if detects["result"]:
                local_name, local_score = get_results(
                    line, scores, psortbs, out_p, seq_name, fuzzy)
            if line.startswith("Final Prediction:"):
                detects["score"] = False
                detects["result"] = True
            if line.startswith("SeqID:"):
                seq_name = line.replace("SeqID: ", "")
                scores = []
            if line.startswith("Localization Scores:"):
                detects["score"] = True
    return psortbs


def print_gff(gffs, psortbs, out_m):
    for gff in gffs:
        detect = False
        for psortb in psortbs:
            if (gff.feature == "CDS") and \
               (gff.start == psortb["start"]) and \
               (gff.end == psortb["end"]) and \
               (gff.strand == psortb["strand"]):
                if "protein_id" in gff.attributes.keys():
                    if gff.attributes["protein_id"] == psortb["protein_id"]:
                        detect = True
                        break
                elif "locus_tag" in gff.attributes.keys():
                    if gff.attributes["locus_tag"] == psortb["protein_id"]:
                        detect = True
                        break
                else:
                    if gff.attributes["ID"] == psortb["protein_id"]:
                        detect = True
                        break
        if detect:
            gff.attribute_string = gff.attribute_string + \
                                   ";subcellular_localization=" + \
                                   psortb["local"]
            out_m.write("\t".join([gff.info_without_attributes,
                                   gff.attribute_string + "\n"]))
        else:
            out_m.write(gff.info + "\n")


def extract_psortb(psortb_table, out_psortb, merge_gff, out_merge, fuzzy):
    '''extract and re-generate the output information of psorb'''
    gffs = []
    if merge_gff:
        if out_merge is None:
            print("Error: Assign a name of output merged annotation file.")
            sys.exit()
        out_m = open(out_merge, "w")
        for entry in Gff3Parser().entries(open(merge_gff)):
            gffs.append(entry)
        gffs = sorted(gffs, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    else:
        out_m = None
    out_p = open(out_psortb, "w")
    out_p.write("#Genome\tProtein\tStrand\tStart\tEnd\tLocation\tScore\n")
    psortbs = get_information(psortb_table, out_p, fuzzy)
    if merge_gff:
        print_gff(gffs, psortbs, out_m)



from copy import deepcopy
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
plt.style.use('ggplot')


def fig(rowlabels, collabels, cells, filename, max_color, min_color):
    row_num = len(rowlabels) / 100
    if row_num == 0:
        row_num = 1
    col_num = len(collabels) / 8
    if col_num == 0:
        col_num = 1
    plt.figure(figsize=(18*col_num, 10*row_num), edgecolor=None)
    img = plt.imshow(cells, interpolation='none', aspect='auto', cmap="RdBu_r")
    plt.xticks(range(len(collabels)), collabels, fontsize=6)
    plt.yticks(range(len(rowlabels)), rowlabels, fontsize=6)
    plt.colorbar(fraction=0.046, pad=0.04)
    img.set_clim(vmin=min_color, vmax=max_color)
    plt.savefig(filename)


def plot_table(plots, max_color, min_color, filename):
    rowlabels = []
    collabels = []
    cells = []
    first = True
    t_num = 0
    for plot in plots:
        for key, value in plot.items():
            rowlabels.append(key)
        cell = []
        for cond, tracks in value.items():
            for track, cover in tracks.items():
                if first:
                    name = track
                    if len(track) > 16:
                        diff = int(len(name) / 16)
                        for i in range(diff):
                            name = (name[:(16)*(i+1)+i] + "\n" +
                                    name[(16)*(i+1)+i:])
                    collabels.append(name)
                cell.append(round(cover, 1))
        cells.append(deepcopy(cell))
        first = False
        if len(rowlabels) >= 500:
            plotname = (filename[:-4] + "_" + str(t_num) + "-" +
                        str(t_num + 500) + ".png")
            fig(rowlabels, collabels, cells, plotname, max_color, min_color)
            t_num = t_num + 500
            rowlabels = []
            cells = []
    if t_num == 0:
        fig(rowlabels, collabels, cells, filename, max_color, min_color)



import os
import sys
import numpy as np
from annogesiclib.lib_reader import read_wig, read_libs
from annogesiclib.coverage_detection import get_repmatch


def check_tex_conds(tracks, libs, texs, check_tex, conds, tex_notex):
    for track in tracks:
        for lib in libs:
            if lib["name"] == track:
                if "tex" in lib["type"]:
                    type_ = "tex"
                else:
                    type_ = "frag"
                index = "_".join([lib["cond"]])
                if len(texs) != 0:
                    for key, num in texs.items():
                        if track in key:
                            if (texs[key] >= tex_notex) and (
                                    key not in check_tex):
                                check_tex.append(key)
                                if index not in conds.keys():
                                    conds[index] = 1
                                else:
                                    conds[index] += 1
                else:
                    if index not in conds.keys():
                        conds[index] = 1
                    else:
                        conds[index] += 1


def detect_hight_toler(cover, height, tmp_covers, tracks, lib_track):
    if cover > height:
        if tmp_covers["best"] < cover:
            tmp_covers["best"] = cover
        tracks.append(lib_track)
    else:
        if cover > tmp_covers["toler"]:
            tmp_covers["toler"] = cover


def elongation(lib_conds, template_texs, libs, strand, trans,
               args_tran, strain, tolers):
    '''check coverage and replicate match to form transcript'''
    first = True
    pre_pos = -1
    check_tex = []
    tracks = []
    conds = {}
    pre_wig = None
    detect = False
    texs = template_texs.copy()
    tmp_covers = {"best": 0, "toler": -1}
    for cond, lib_tracks in lib_conds.items():
        for lib_name, covers in lib_tracks.items():
            index_pos = 0
            for cover in covers:
                for cond, lib_tracks in lib_conds.items():
                    for lib_track in lib_tracks.keys():
                        real_track = lib_track.split("|")[-3]
                        if index_pos < len(lib_tracks[lib_track]):
                            compare_cover = lib_tracks[lib_track][index_pos]
                        else:
                            compare_cover = 0
                        detect_hight_toler(
                                compare_cover, args_tran.height,
                                tmp_covers, tracks, real_track)
                for track in tracks:
                    if len(texs) != 0:
                        for key, num in texs.items():
                            if track in key:
                                texs[key] += 1
                check_tex_conds(tracks, libs, texs, check_tex,
                                conds, args_tran.tex)
                for cond, detect_num in conds.items():
                    if ("tex" in cond):
                        tex_rep = get_repmatch(args_tran.replicates["tex"], cond)
                        if detect_num >= tex_rep:
                            detect = True
                    elif ("frag" in cond):
                        frag_rep = get_repmatch(args_tran.replicates["frag"], cond)
                        if detect_num >= frag_rep:
                            detect = True
                if detect:
                    detect = False
                    trans[strain].append(tmp_covers["best"])
                else:
                    trans[strain].append(-1)
                if (tmp_covers["toler"] != -1):
                    tolers.append(tmp_covers["toler"])
                else:
                    tolers.append(args_tran.height + 10)
                tmp_covers = {"best": 0, "toler": -1}
                tracks = []
                conds = {}
                check_tex = []
                texs = template_texs.copy()
                index_pos += 1
            break
        break


def transfer_to_tran(wigs, libs, template_texs, strand, args_tran):
    '''check coverage and replicate match to form transcript'''
    tolers = {}
    trans = {}
    detect = False
    for strain, lib_conds in wigs.items():
        if strain not in trans:
            trans[strain] = []
            tolers[strain] = []
        elongation(lib_conds, template_texs, libs, strand, trans,
                   args_tran, strain, tolers[strain])
    return tolers, trans


def print_transcript(finals, out):
    for strain, datas in finals.items():
        num = 0
        datas = sorted(datas, key=lambda x: (
            x["start"], x["end"], x["strand"]))
        for data in datas:
            name = '%0*d' % (5, num)
            attribute = ";".join(["=".join(items) for items in ([
                        ("ID", strain + "_transcript" + str(num)),
                        ("Name", "transcript_" + name),
                        ("high_coverage", str(data["high"])),
                        ("low_coverage", str(data["low"])),
                        ("detect_lib", data["wig"])])])
            out.write("\t".join([str(field) for field in [
                strain, "ANNOgesic", "transcript", str(data["start"]),
                str(data["end"]), ".", data["strand"], ".",
                attribute]]) + "\n")
            num += 1


def fill_gap_and_print(trans, strand, finals, tolers, wig_type, args_tran):
    '''compare transcript with CDS to modify transcript(merge mutliple 
    transcript based on overlap with the same CDS)'''
    for strain, covers in trans.items():
        if strain not in finals:
            finals[strain] = []
        first = True
        start = -1
        end = -1
        pre_cover = None
        cover_pos = 1
        for cover in covers:
            fit = True
            if cover != -1:
                if first:
                    first = False
                    start = cover_pos
                    high_cover = cover
                    low_cover = cover
                else:
                    if (cover_pos - pre_pos) <= args_tran.tolerance:
                        if cover_pos - pre_pos > 1:
                            for toler_strain, toler_datas in tolers.items():
                                if toler_strain == strain:
                                    toler_covers = toler_datas[
                                            (pre_pos - 1): cover_pos]
                                    for toler_cover in toler_covers:
                                        if (toler_cover < args_tran.low_cutoff):
                                            fit = False
                                            break
                        if fit:
                            end = cover_pos
                            if high_cover < cover:
                                high_cover = cover
                            if low_cover > cover:
                                low_cover = cover
                    if ((cover_pos - pre_pos) >
                            args_tran.tolerance) or (not fit):
                        if (start != -1) and (end != -1) and (
                                (end - start) >= args_tran.width):
                            finals[strain].append({
                                "start": start, "end": end, "strand": strand,
                                "high": high_cover, "low": low_cover,
                                "wig": wig_type})
                        start = cover_pos
                        end = -1
                        high_cover = cover
                        low_cover = cover
                pre_cover = cover
                pre_pos = cover_pos
            cover_pos += 1
        if (len(covers) != 0) and (not first) and (
                (start != -1) and (end != -1) and (
                (end - start) >= args_tran.width)):
            finals[strain].append({
                "start": start, "end": end, "strand": strand,
                "high": high_cover, "low": low_cover,
                "wig": wig_type})
    return finals

def detect_transcript(wig_f_file, wig_r_file, wig_folder, input_lib,
                      out_file, wig_type, args_tran):
    out = open(out_file, "w")
    out.write("##gff-version 3\n")
    finals = {}
    libs, texs = read_libs(input_lib, wig_folder)
    wig_fs = read_wig(wig_f_file, "+", libs)
    wig_rs = read_wig(wig_r_file, "-", libs)
    tolers_f, tran_fs = transfer_to_tran(wig_fs, libs, texs, "+", args_tran)
    tolers_r, tran_rs = transfer_to_tran(wig_rs, libs, texs, "-", args_tran)
    fill_gap_and_print(tran_fs, "+", finals, tolers_f, wig_type, args_tran)
    fill_gap_and_print(tran_rs, "-", finals, tolers_r, wig_type, args_tran)
    print_transcript(finals, out)
    out.close()
    del wig_fs
    del wig_rs



import os, gc
import math
import numpy as np
from annogesiclib.gff3 import Gff3Parser
from annogesiclib.lib_reader import read_wig, read_libs
from annogesiclib.coverage_detection import coverage_comparison, get_repmatch
from annogesiclib.coverage_detection import replicate_comparison
from annogesiclib.args_container import ArgsContainer


def import_data(strand, strain, pos, utr, type_, name, srna_cover, pro):
    if type_ == "TSS":
        data = {"strand": strand, "strain": strain, "start": pos["start"],
                "end": pos["end"], "utr": utr, "start_tss": name,
                "end_cleavage": pro, "start_cleavage": "NA",
                "datas": srna_cover}
    elif type_ == "cleavage":
        data = {"strand": strand, "strain": strain, "start": pos["start"],
                "end": pos["end"], "utr": utr, "start_tss": "NA",
                "start_cleavage": name, "datas": srna_cover,
                "end_cleavage": pro}
    else:
        data = {"strand": strand, "strain": strain, "start": pos["start"],
                "end": pos["end"], "utr": utr, "start_tss": "NA",
                "start_cleavage": "NA", "datas": srna_cover,
                "end_cleavage": pro}
    del(srna_cover)
    return data


def import_inter(strand, strain, start, end, length):
    return {"strand": strand, "strain": strain,
            "start": start, "end": end, "len_CDS": length}


def get_terminal(cdss, inters, seq, type_):
    first_p = True
    first_m = True
    pre_strain = ""
    if type_ == "start":
        for cds in cdss:
            if cds.seq_id != pre_strain:
                pre_strain = cds.seq_id
                first_p = True
                first_m = True
            if (cds.strand == "+") and (first_p):
                first_p = False
                inters.append(import_inter(cds.strand, cds.seq_id,
                                           1, cds.start, 0))
            elif (cds.strand == "-") and (first_m):
                first_m = False
                inters.append(import_inter(cds.strand, cds.seq_id,
                                           1, cds.start, 0))
    elif type_ == "end":
        for cds in reversed(cdss):
            if cds.seq_id != pre_strain:
                pre_strain = cds.seq_id
                first_p = True
                first_m = True
            if (cds.strand == "+") and (first_p):
                first_p = False
                inters.append(import_inter(
                    cds.strand, cds.seq_id, cds.end,
                    len(seq[cds.seq_id]), cds.end - cds.start))
            elif (cds.strand == "-") and (first_m):
                first_m = False
                inters.append(import_inter(
                    cds.strand, cds.seq_id, cds.start,
                    len(seq[cds.seq_id]), cds.end - cds.start))


def check_pos(cover, check_point, checks, cover_pos):
    if (cover_pos >= min(check_point["utr_start"],
                            check_point["utr_end"])) and (
        cover_pos <= max(check_point["utr_start"],
                            check_point["utr_end"])):
        checks["utr"] = True
    if (cover_pos >= min(check_point["srna_start"],
                            check_point["srna_end"])) and (
        cover_pos <= max(check_point["srna_start"],
                            check_point["srna_end"])):
        checks["srna"] = True


def check_start_and_end(start, end, covers, strand, fuzzy):
    if strand == "-":
        if (start - 2 - fuzzy) < 0:
            c_start = 0
        else:
            c_start = start - 2 - fuzzy
        if (end + 1) > len(covers):
            c_end = len(covers)
        else:
            c_end = end + 1
    else:
        if (start - 2) < 0:
            c_start = 0
        else:
            c_start = start - 2
        if (end + 1 + fuzzy) > len(covers):
            c_end = len(covers)
        else:
            c_end = end + 1 + fuzzy
    return c_start, c_end


def set_cover_and_point(cover_results, inter, covers, pos, fuzzy_end):
    check_point = {"srna_start": 0, "srna_end": 0,
                   "utr_start": 0, "utr_end": 0}
    if pos["ori_start"] - 2 < 0:
        ori_start = 2
    else:
        ori_start = pos["ori_start"]
    if inter["strand"] == "-":
        c_start, c_end = check_start_and_end(ori_start, pos["ori_end"],
                                             covers, "-", fuzzy_end)
        covers = covers[c_start: c_end]
        covers = covers[::-1]
        check_point["srna_start"] = pos["end"] + 1
        check_point["srna_end"] = pos["start"] - 2 - fuzzy_end
        check_point["utr_start"] = pos["ori_end"]
        check_point["utr_end"] = ori_start
        start = c_start
        end = c_end
    elif inter["strand"] == "+":
        c_start, c_end = check_start_and_end(ori_start, pos["ori_end"], 
                                             covers, "+", fuzzy_end)
        covers = covers[c_start: c_end]
        check_point["srna_start"] = pos["start"] - 2
        check_point["srna_end"] = pos["end"] + 1 + fuzzy_end
        check_point["utr_start"] = ori_start
        check_point["utr_end"] = pos["ori_end"]
        start = c_start
        end = c_end
    cover_results["check_point"] = check_point
    cover_results["covers"] = covers
    return start, end


def detect_cover_utr_srna(cover_results, pos, inter, cond, track,
                          args_srna, lib_type, start, end, strand):
    datas = {"num": 0, "cover_tmp": {"5utr": 0, "total": 0, "ori_total": 0},
             "checks": {"first": True, "detect_decrease": False,
                        "srna": False, "utr": False},
             "final_poss": {"start": pos["start"], "end": pos["end"]}}
    index_pos = 0
    for cover in cover_results["covers"]:
        if strand == "+":
            cover_pos = start + index_pos
        else:
            cover_pos = end - index_pos
        datas["checks"]["srna"] = False
        datas["checks"]["utr"] = False
        check_pos(cover, cover_results["check_point"], datas["checks"],
                  cover_pos)
        if datas["checks"]["utr"]:
            datas["cover_tmp"]["ori_total"] = \
                datas["cover_tmp"]["ori_total"] + cover
        if datas["checks"]["srna"]:
            datas["cover_tmp"]["total"] = \
                datas["cover_tmp"]["total"] + cover
            datas["checks"]["first"] = coverage_comparison(
                    cover, cover_results["cover_sets"], cover_results["pos"],
                    datas["checks"]["first"], inter["strand"], cover_pos)
            if (datas["checks"]["first"] is not True) and (
                    cover_results["cover_sets"]["high"] > 0):
                if (cover_results["type"] == "5utr") or (
                        cover_results["type"] == "3utr") or (
                        (cover_results["type"] == "interCDS") and (
                         cover_results["intercds"] == "TSS")):
                    if ((cover_results["cover_sets"]["low"] /
                            cover_results["cover_sets"]["high"]) <
                            args_srna.decrease_utr) and (
                            cover_results["cover_sets"]["low"] > -1):
                        datas["checks"]["detect_decrease"] = True
                        datas["cover_tmp"]["5utr"] = cover
            if datas["checks"]["detect_decrease"]:
                go_out = get_cover_5utr(datas, cover_results["cover_sets"],
                                        cover, inter, args_srna, cover_pos)
                if go_out is True:
                    break
        index_pos += 1
    if (datas["checks"]["first"] is not True) and (
            cover_results["cover_sets"]["high"] > 0):
        check_import_srna_covers(datas, cover_results, inter, cond, track,
                                 cover, pos, args_srna, lib_type)


def get_coverage(wigs, inter, pos, type_, intercds_type, args_srna):
    cover_results = {"srna_covers": {}, "utr_covers": {},
                     "cover_sets": {"high": 0, "low": 0, "best": -1},
                     "pos": {"low": 0, "high": 0},
                     "type": type_, "intercds": intercds_type}
    for wig_strain, conds in wigs.items():
        if wig_strain == inter["strain"]:
            for cond, tracks in conds.items():
                cover_results["srna_covers"][cond] = []
                cover_results["utr_covers"][cond] = []
                for lib_name, covers in tracks.items():
                    track = lib_name.split("|")[-3]
                    lib_strand = lib_name.split("|")[-2]
                    lib_type = lib_name.split("|")[-1]
                    start, end = set_cover_and_point(
                            cover_results, inter, covers,
                            pos, args_srna.fuzzy_utr)
                    detect_cover_utr_srna(cover_results, pos, inter,
                                          cond, track, args_srna, lib_type,
                                          start, end, lib_strand)
    return cover_results["srna_covers"], cover_results["utr_covers"]


def get_cover_5utr(datas, cover_sets, cover, inter, args_srna, cover_pos):
    go_out = False
    if (datas["num"] == args_srna.fuzzy_utr) or (
            datas["cover_tmp"]["5utr"] == 0) or (
            (cover > datas["cover_tmp"]["5utr"]) and (
            cover / datas["cover_tmp"]["5utr"]) > (
            1 + args_srna.decrease_utr)):
        if inter["strand"] == "+":
            datas["final_poss"]["end"] = cover_pos
        elif inter["strand"] == "-":
            datas["final_poss"]["start"] = cover_pos
        go_out = True
    elif (cover <= datas["cover_tmp"]["5utr"]):
        if (cover / datas["cover_tmp"]["5utr"]) >= (
                args_srna.decrease_utr / 2):
            datas["num"] += 1
        else:
            datas["num"] = 0
        datas["cover_tmp"]["5utr"] = cover
        cover_sets["low"] = cover
    elif (cover > datas["cover_tmp"]["5utr"]) and (
            (cover / datas["cover_tmp"]["5utr"]) <= (
            1 + args_srna.decrease_utr)):
        datas["num"] += 1
    return go_out


def import_cover(inter, covers, track, cover_sets, avgs, cover,
                 final_poss, pos, lib_type):
    if final_poss["start"] < final_poss["end"]:
        if (inter["strand"] == "+") and (final_poss["end"] > pos["end"]):
            final_poss["end"] = pos["end"]
        elif (inter["strand"] == "-") and (final_poss["start"] < pos["start"]):
            final_poss["start"] = pos["start"]
        covers.append({"track": track,
                       "high": cover_sets["high"],
                       "low": cover_sets["low"], "avg": avgs["avg"],
                       "type": lib_type,
                       "ori_avg": avgs["ori_avg"],
                       "final_start": final_poss["start"],
                       "final_end": final_poss["end"]})


def check_import_srna_covers(datas, cover_results, inter, cond, track,
                             cover, pos, args_srna, lib_type):
    avgs = {"avg": datas["cover_tmp"]["total"] / float(
                   pos["end"] - pos["start"] + 1),
            "ori_avg": datas["cover_tmp"]["ori_total"] / float(
                       pos["ori_end"] - pos["ori_start"] + 1)}
    if ((cover_results["type"] == "5utr") and (
            (cover_results["intercds"] == "tsspro") or (
             datas["checks"]["detect_decrease"]))) or (
        (cover_results["type"] == "3utr") and (
             cover_results["intercds"] == "two_pro")) or (
        (cover_results["type"] == "3utr") and (
             cover_results["intercds"] != "two_pro") and (
             datas["checks"]["detect_decrease"])) or (
        (cover_results["type"] == "3utr") and (
             cover_results["intercds"] != "two_pro") and (
             (pos["end"] - pos["start"]) >= args_srna.min_len) and (
             (pos["end"] - pos["start"]) <= args_srna.max_len)) or (
        (cover_results["type"] == "interCDS") and (
             cover_results["intercds"] == "TSS") and (
             datas["checks"]["detect_decrease"])) or (
        (cover_results["type"] == "interCDS") and (
            (cover_results["intercds"] == "tss_pro") or (
             cover_results["intercds"] == "two_pro"))):
        import_cover(inter, cover_results["srna_covers"][cond], track,
                     cover_results["cover_sets"], avgs, cover,
                     datas["final_poss"], pos, lib_type)
        import_cover(inter, cover_results["utr_covers"][cond], track,
                     cover_results["cover_sets"], avgs, cover,
                     datas["final_poss"], pos, lib_type)
    else:
        import_cover(inter, cover_results["utr_covers"][cond], track,
                     cover_results["cover_sets"], avgs, cover,
                     datas["final_poss"], pos, lib_type)


def detect_3utr_pro(inter, pos, wigs, utr_type, args_srna):
    '''3UTR start with processing site'''
    for pro in args_srna.pros:
        if (pro.seq_id == inter["strain"]) and (
                pro.strand == inter["strand"]):
            if (pro.start >= pos["start"]) and (pro.start <= pos["end"]):
                if pro.strand == "+":
                    if ((pos["end"] - pro.start) >= args_srna.min_len) and (
                            (pos["end"] - pro.start) <= args_srna.max_len):
                        n_pos = import_position(
                                pro.start, pos["end"],
                                pos["ori_start"], pos["ori_end"])
                        srna_covers, utr_covers = get_coverage(
                                wigs, inter, n_pos, utr_type, "pro", args_srna)
                        args_srna.utrs.append(import_data(
                            inter["strand"], inter["strain"], n_pos, utr_type,
                            "NA", "NA", utr_covers, "NA"))
                        args_srna.srnas.append(import_data(
                            inter["strand"], inter["strain"], n_pos, "3utr",
                            "cleavage", "Cleavage:" + "_".join([
                                str(pro.start), pro.strand]),
                            srna_covers, "NA"))
                    elif (pos["end"] - pro.start) > args_srna.max_len:
                        detect_twopro(inter, pos, wigs, utr_type,
                                      "3utr", args_srna)
                else:
                    if ((pro.start - pos["start"]) >= args_srna.min_len) and (
                            (pro.start - pos["start"]) <= args_srna.max_len):
                        n_pos = import_position(
                                pos["start"], pro.start,
                                pos["ori_start"], pos["ori_end"])
                        srna_covers, utr_covers = get_coverage(
                                wigs, inter, n_pos, utr_type, "pro", args_srna)
                        args_srna.utrs.append(import_data(
                            inter["strand"], inter["strain"], n_pos, utr_type,
                            "NA", "NA", utr_covers, "NA"))
                        args_srna.srnas.append(import_data(
                            inter["strand"], inter["strain"], n_pos, "3utr",
                            "cleavage", "Cleavage:" + "_".join([
                                str(pro.start), pro.strand]),
                            srna_covers, "NA"))
                    elif (pro.start - pos["start"]) > args_srna.max_len:
                        detect_twopro(inter, pos, wigs, utr_type,
                                      "3utr", args_srna)
            if (pro.start > pos["end"] + args_srna.fuzzy_tsss[utr_type]):
                break


def detect_twopro(inter, pos, wigs, utr_type, import_type, args_srna):
    '''the sRNA is associated with two processing sites'''
    pros = []
    for pro in args_srna.pros:
        if (pro.seq_id == inter["strain"]) and (
                pro.strand == inter["strand"]):
            if (pro.start >= pos["start"]) and (pro.start <= pos["end"]):
                pros.append(pro)
    first = True
    pre_pro = None
    for pro in pros:
        if first:
            first = False
        else:
            if ((pro.start - pre_pro.start) >= args_srna.min_len) and (
                    (pro.start - pre_pro.start) <= args_srna.max_len):
                n_pos = import_position(pre_pro.start, pro.start,
                                        pos["ori_start"], pos["ori_end"])
                srna_covers, utr_covers = get_coverage(
                        wigs, inter, n_pos, utr_type, "two_pro", args_srna)
                args_srna.utrs.append(import_data(
                    inter["strand"], inter["strain"], n_pos, utr_type, "NA",
                    "NA", utr_covers, "Cleavage:" + "_".join(
                        [str(pro.start), pro.strand])))
                args_srna.srnas.append(import_data(
                    inter["strand"], inter["strain"], n_pos, import_type,
                    "cleavage", "Cleavage:" + "_".join(
                        [str(pre_pro.start), pro.strand]),
                    srna_covers, "Cleavage:" + "_".join(
                        [str(pro.start), pro.strand])))
        pre_pro = pro


def decrease_pos(covers, pos, strand):
    longer = -1
    for cond, datas in covers.items():
        for data in datas:
            for key, value in data.items():
                if key == pos:
                    if longer == -1:
                        longer = value
                    elif (strand == "+") and (value > longer) and (
                            longer != -1):
                        longer = value
                    elif (strand == "-") and (value < longer) and (
                            longer != -1):
                        longer = value
    return longer


def get_decrease(inter, wigs, tss, pos, utr_type, args_srna):
    '''check the coverage decrease'''
    if inter["strand"] == "+":
        n_pos = import_position(tss.start, pos["end"], pos["ori_start"],
                                pos["ori_end"])
        srna_covers, utr_covers = get_coverage(
                wigs, inter, n_pos, utr_type, "TSS", args_srna)
        utr_pos = decrease_pos(utr_covers, "final_end", "+")
        n_pos = import_position(tss.start, utr_pos, pos["ori_start"],
                                pos["ori_end"])
        args_srna.utrs.append(import_data(
            inter["strand"], inter["strain"], n_pos, utr_type, "NA", "NA",
            utr_covers, "NA"))
        if len(srna_covers) != 0:
            srna_pos = decrease_pos(srna_covers, "final_end", "+")
            n_pos = import_position(tss.start, srna_pos, pos["ori_start"],
                                    pos["ori_end"])
            args_srna.srnas.append(import_data(
                inter["strand"], inter["strain"], n_pos, utr_type, "TSS",
                "TSS:" + "_".join([str(tss.start), tss.strand]),
                srna_covers, "NA"))
    else:
        n_pos = import_position(pos["start"], tss.start, pos["ori_start"],
                                pos["ori_end"])
        srna_covers, utr_covers = get_coverage(
                wigs, inter, n_pos, utr_type, "TSS", args_srna)
        utr_pos = decrease_pos(utr_covers, "final_start", "-")
        n_pos = import_position(utr_pos, tss.start, pos["ori_start"],
                                pos["ori_end"])
        args_srna.utrs.append(import_data(
            inter["strand"], inter["strain"], n_pos, utr_type, "NA", "NA",
            utr_covers, "NA"))
        if len(srna_covers) != 0:
            srna_pos = decrease_pos(srna_covers, "final_start", "-")
            n_pos = import_position(srna_pos, tss.start, pos["ori_start"],
                                    pos["ori_end"])
            args_srna.srnas.append(import_data(
                inter["strand"], inter["strain"], n_pos, utr_type, "TSS",
                "TSS:" + "_".join([str(tss.start), tss.strand]),
                srna_covers, "NA"))


def import_append_normal(inter, tss, pro, pos, wigs, utr_type, args_srna):
    if inter["strand"] == "+":
        n_pos = import_position(tss.start, pro.start,
                                pos["ori_start"], pos["ori_end"])
        srna_covers, utr_covers = get_coverage(
                wigs, inter, n_pos, utr_type, "tsspro", args_srna)
        args_srna.utrs.append(import_data(
            inter["strand"], inter["strain"], n_pos, utr_type, "NA", "NA",
            utr_covers, "Cleavage:" + "_".join([str(pro.start), pro.strand])))
        args_srna.srnas.append(import_data(
            inter["strand"], inter["strain"], n_pos, utr_type, "TSS",
            "TSS:" + "_".join([str(tss.start), tss.strand]), srna_covers,
            "Cleavage:" + "_".join([str(pro.start), pro.strand])))
    else:
        n_pos = import_position(pro.start, tss.start,
                                pos["ori_start"], pos["ori_end"])
        srna_covers, utr_covers = get_coverage(
                wigs, inter, n_pos, utr_type, "tsspro", args_srna)
        args_srna.utrs.append(import_data(
            inter["strand"], inter["strain"], n_pos, utr_type, "NA", "NA",
            utr_covers, "Cleavage:" + "_".join([str(pro.start), pro.strand])))
        args_srna.srnas.append(import_data(
            inter["strand"], inter["strain"], n_pos, utr_type, "TSS",
            "TSS:" + "_".join([str(tss.start), tss.strand]), srna_covers,
            "Cleavage:" + "_".join([str(pro.start), pro.strand])))


def detect_normal(diff, wigs, inter, pos, utr_type, tss, args_srna):
    '''normal case, UTR-derived sRNA with TSS'''
    if (diff >= args_srna.min_len) and (
            diff <= args_srna.max_len):
        srna_covers, utr_covers = get_coverage(
                wigs, inter, pos, utr_type, "TSS", args_srna)
        args_srna.utrs.append(import_data(
            inter["strand"], inter["strain"], pos, utr_type, "NA", "NA",
            utr_covers, "NA"))
        args_srna.srnas.append(import_data(
            inter["strand"], inter["strain"], pos, utr_type, "TSS",
            "TSS:" + "_".join([str(tss.start), tss.strand]),
            srna_covers, "NA"))
    elif (diff > args_srna.max_len) and (len(args_srna.pros) != 0):
        detect = False
        for pro in args_srna.pros:
            if (pro.seq_id == inter["strain"]) and (
                    pro.strand == inter["strand"]):
                if (pro.start >= pos["start"]) and (pro.start <= pos["end"]):
                    if ((pro.start - pos["start"]) >= args_srna.min_len) and (
                            (pro.start - pos["start"]) <= args_srna.max_len):
                        detect = True
                        import_append_normal(inter, tss, pro, pos,
                                             wigs, utr_type, args_srna)
        if not detect:
            get_decrease(inter, wigs, tss, pos, utr_type, args_srna)


def import_position(start, end, ori_start, ori_end):
    return {"start": start, "end": end,
            "ori_start": ori_start, "ori_end": ori_end}


def detect_utr(start, end, inter, utr_type, wigs, args_srna):
    ori_fuzzy = args_srna.fuzzy_tsss[utr_type]
    if "p_" in args_srna.fuzzy_tsss[utr_type]:
        per = float(args_srna.fuzzy_tsss[utr_type].split("_")[-1])
        args_srna.fuzzy_tsss[utr_type] = inter["len_CDS"]*per
    elif "n_" in args_srna.fuzzy_tsss[utr_type]:
        args_srna.fuzzy_tsss[utr_type] = float(
            args_srna.fuzzy_tsss[utr_type].split("_")[-1])
    for tss in args_srna.tsss:
        if (tss.seq_id == inter["strain"]) and (
                tss.strand == inter["strand"]):
            if tss.strand == "+":
                start_fuzzy = start - args_srna.fuzzy_tsss[utr_type]
                if (tss.start >= start_fuzzy) and (tss.start <= end):
                    n_pos = import_position(tss.start, end, start, end)
                    detect_normal(end - tss.start, wigs, inter, n_pos,
                                  utr_type, tss, args_srna)
                elif tss.start > end:
                    break
            else:
                end_fuzzy = end + args_srna.fuzzy_tsss[utr_type]
                if (tss.start >= start) and (tss.start <= end_fuzzy):
                    n_pos = import_position(start, tss.start, start, end)
                    detect_normal(tss.start - start, wigs, inter, n_pos,
                                  utr_type, tss, args_srna)
                if tss.start > end_fuzzy:
                    break
    if (utr_type == "3utr") and (len(args_srna.pros) != 0):
        pos = import_position(start, end, start, end)
        detect_3utr_pro(inter, pos, wigs, utr_type, args_srna)
    if (utr_type == "interCDS") and (len(args_srna.pros) != 0):
        pos = import_position(start, end, start, end)
        detect_twopro(inter, pos, wigs, utr_type, "interCDS", args_srna)
    args_srna.fuzzy_tsss[utr_type] = ori_fuzzy


def run_utr_detection(wigs, inter, start, end, utr_type, args_srna):
    if end - start >= 0:
        if (utr_type == "3utr") or (
                utr_type == "5utr") or (
                utr_type == "interCDS"):
            detect_utr(start, end, inter, utr_type, wigs, args_srna)
        else:
            pos = import_position(start, end, start, end)
            srna_covers, utr_covers = get_coverage(
                    wigs, inter, pos, utr_type, "NA", args_srna)
            args_srna.utrs.append(import_data(
                inter["strand"], inter["strain"],
                pos, utr_type, "NA", "NA", utr_covers, "NA"))


def class_utr(inter, ta, args_srna, wig_fs, wig_rs):
    '''classify the UTR-dervied sRNA'''
    if inter["strand"] == "+":
        if (inter["start"] <= ta.end) and (
                inter["end"] >= ta.end) and (
                ta.start <= inter["start"]):
            run_utr_detection(wig_fs, inter, inter["start"] + 1,
                              ta.end, "3utr", args_srna)
        elif (inter["start"] <= ta.start) and (
                inter["end"] >= ta.start) and (
                ta.end >= inter["end"]):
            run_utr_detection(wig_fs, inter, ta.start,
                              inter["end"] - 1, "5utr", args_srna)
        elif (inter["start"] >= ta.start) and (inter["end"] <= ta.end):
            run_utr_detection(wig_fs, inter, inter["start"] + 1,
                              inter["end"] - 1, "interCDS", args_srna)
    else:
        if (inter["start"] <= ta.end) and (
                inter["end"] >= ta.end) and (
                ta.start <= inter["start"]):
            run_utr_detection(wig_rs, inter, inter["start"] + 1,
                              ta.end, "5utr", args_srna)
        elif (inter["start"] <= ta.start) and (
                inter["end"] >= ta.start) and (
                ta.end >= inter["end"]):
            run_utr_detection(wig_rs, inter, ta.start,
                              inter["end"] - 1, "3utr", args_srna)
        elif (inter["start"] >= ta.start) and (inter["end"] <= ta.end):
            run_utr_detection(wig_rs, inter,  inter["start"] + 1,
                              inter["end"] - 1, "interCDS", args_srna)


def median_score(lst, per):
    '''if the cutoff is assigned by precentage, 
    it can get the corresponding number'''
    sortedLst = sorted(lst)
    lstLen = len(lst)
    index = int((lstLen - 1) * per)
    if lstLen != 0:
        return sortedLst[index]
    else:
        return 0


def print_file(num, srna, start, end, srna_datas, args_srna):
    name = '%0*d' % (5, num)
    args_srna.out_t.write(
            "{0}\t{1}\t{2}\t{3}\t{4}\t{5}\t{6}\t{7}\t{8}\t{9}\t".format(
                srna["strain"], name, start, end, srna["strand"],
                ";".join(srna_datas["conds"].keys()),
                ";".join(srna_datas["conds"].values()),
                srna_datas["best"], srna_datas["high"], srna_datas["low"]))
    attribute_string = ";".join(
        ["=".join(items) for items in [
            ["ID", srna["strain"] + "_srna_utr" + str(num)],
            ["Name", "UTR_sRNA_" + name],
            ["sRNA_type", srna["utr"]],
            ["best_avg_coverage", str(srna_datas["best"])],
            ["best_high_coverage", str(srna_datas["high"])],
            ["best_low_coverage", str(srna_datas["low"])],
            ["with_TSS", srna["start_tss"]],
            ["start_cleavage", srna["start_cleavage"]],
            ["end_cleavage", srna["end_cleavage"]]]])
    args_srna.out.write("\t".join([str(field) for field in [
                        srna["strain"], "ANNOgesic", "ncRNA", str(start),
                        str(end), ".", srna["strand"], ".",
                        attribute_string]]) + "\n")
    if not args_srna.table_best:
        first = True
        for data in srna_datas["detail"]:
            if first:
                args_srna.out_t.write("{0}(avg={1};high={2};low={3})".format(
                                      data["track"], data["avg"],
                                      data["high"], data["low"]))
                first = False
            else:
                args_srna.out_t.write(";{0}(avg={1};high={2};low={3})".format(
                                      data["track"], data["avg"],
                                      data["high"], data["low"]))
    else:
        args_srna.out_t.write("{0}(avg={1};high={2};low={3})".format(
                              srna_datas["track"], srna_datas["best"],
                              srna_datas["high"], srna_datas["low"]))
    args_srna.out_t.write("\n")


def detect_srna(median, args_srna):
    '''check the sRNA candidates and print it out'''
    num = 0
    if len(args_srna.srnas) != 0:
        for srna in args_srna.srnas:
            if srna["strain"] in median.keys():
                srna_datas = replicate_comparison(
                        args_srna, srna["datas"],
                        srna["strand"], "sRNA_utr_derived",
                        median[srna["strain"]][srna["utr"]],
                        args_srna.coverages, srna["utr"], None, None,
                        args_srna.texs)
                if srna_datas["best"] != 0:
                    if (srna["utr"] == "5utr") or (
                            srna["utr"] == "interCDS"):
                        start = srna_datas["start"]
                        end = srna_datas["end"]
                    elif srna["utr"] == "3utr":
                        start = srna["start"]
                        end = srna["end"]
                    if (math.fabs(start - end) >= args_srna.min_len) and (
                            math.fabs(start - end) <= args_srna.max_len):
                        print_file(num, srna, start, end,
                                   srna_datas, args_srna)
                        num += 1


def read_data(args_srna):
    cdss = []
    tas = []
    tsss = []
    pros = []
    seq = {}
    gff_parser = Gff3Parser()
    fh = open(args_srna.gff_file, "r")
    for entry in gff_parser.entries(fh):
        if (entry.feature == "CDS") or (
                entry.feature == "tRNA") or (
                entry.feature == "rRNA"):
            if ("product" in entry.attributes.keys()) and (args_srna.hypo):
                if "hypothetical protein" not in entry.attributes["product"]:
                    cdss.append(entry)
            else:
                cdss.append(entry)
    fh.close()
    fh = open(args_srna.ta_file, "r")
    for entry in gff_parser.entries(fh):
        tas.append(entry)
    fh.close()
    fh = open(args_srna.tss_file, "r")
    for entry in gff_parser.entries(fh):
        tsss.append(entry)
    fh.close()
    if args_srna.pro_file is not None:
        fh = open(args_srna.pro_file, "r")
        for entry in gff_parser.entries(fh):
            pros.append(entry)
        fh.close()
    with open(args_srna.seq_file, "r") as s_f:
        for line in s_f:
            line = line.strip()
            if line.startswith(">"):
                strain = line[1:]
                seq[strain] = ""
            else:
                seq[strain] = seq[strain] + line
    cdss = sorted(cdss, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    tas = sorted(tas, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    tsss = sorted(tsss, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    if len(pros) != 0:
        pros = sorted(pros, key=lambda k: (k.seq_id, k.start, k.end, k.strand))
    return cdss, tas, tsss, pros, seq


def get_utr_coverage(utrs):
    covers = {}
    for utr in utrs:
        if utr["strain"] not in covers.keys():
            covers[utr["strain"]] = {"3utr": {}, "5utr": {},
                                     "interCDS": {}}
        for cond, utr_covers in utr["datas"].items():
            for cover in utr_covers:
                if (cover["track"] not in
                        covers[utr["strain"]][utr["utr"]].keys()):
                    covers[utr["strain"]][utr["utr"]][cover["track"]] = []
                covers[utr["strain"]][utr["utr"]][cover["track"]].append(
                        cover["ori_avg"])
    return covers


def get_inter(cdss, inters):
    '''get the intergenic region'''
    for cds1 in cdss:
        for cds2 in cdss:
            if (cds1.seq_id == cds2.seq_id) and \
               (cds1.strand == cds2.strand):
                if (cds2.start > cds1.start):
                    if cds2.start - cds1.end > 1:
                        if cds1.strand == "+":
                            length = cds1.end - cds1.start
                        else:
                            length = cds2.end - cds2.start
                        inters.append(import_inter(cds1.strand, cds1.seq_id,
                                      cds1.end, cds2.start, length))
                    break


def mean_score(lst):
    total = 0
    for li in lst:
        total = total + li
    if len(lst) != 0:
        return (total / len(lst))
    else:
        return 0


def get_utr_cutoff(coverage, mediandict, avgs, strain, utr, track):
    if "n_" in coverage:
        cutoff = float(coverage.split("_")[-1])
        mediandict[strain][utr][track] = {"median": cutoff,
                                          "mean": mean_score(avgs)}
    elif "p_" in coverage:
        cutoff = float(coverage.split("_")[-1])
        mediandict[strain][utr][track] = {"median": median_score(avgs, cutoff),
                                          "mean": mean_score(avgs)}


def set_cutoff(covers, args_srna):
    '''set the cutoff based on the types of sRNA'''
    mediandict = {}
    for strain, utrs in covers.items():
        mediandict[strain] = {"3utr": {}, "5utr": {}, "interCDS": {}}
        for utr, tracks in utrs.items():
            if (utr == "3utr") or (utr == "5utr") or (utr == "interCDS"):
                for track, avgs in tracks.items():
                    if track not in mediandict[strain][utr].keys():
                        mediandict[strain][utr][track] = {}
                    if args_srna.cover_notex is not None:
                        for keys in args_srna.texs.keys():
                            tracks = keys.split("@AND@")
                            if tracks[0] == track:
                                get_utr_cutoff(args_srna.coverages[utr],
                                               mediandict,
                                               avgs, strain, utr, track)
                                break
                            elif tracks[1] == track:
                                get_utr_cutoff(args_srna.cover_notex[utr],
                                               mediandict,
                                               avgs, strain, utr, track)
                                break
                    else:
                        get_utr_cutoff(args_srna.coverages[utr], mediandict,
                                       avgs, strain, utr, track)
    return mediandict


def print_median(out_folder, mediandict):
    '''print the cutoff based on the types of sRNA'''
    out = open(os.path.join(out_folder, "tmp_median"), "a")
    for strain, utrs in mediandict.items():
        for utr, tracks in utrs.items():
            for track, value in tracks.items():
                out.write("\t".join([strain, utr, track,
                                     str(value["median"])]) + "\n")
    out.close()


def free_memory(paras):
    for data in (paras):
        del(data)
    gc.collect()


def utr_derived_srna(args_srna, libs, texs, wig_fs, wig_rs):
    inters = []
    cdss, tas, tsss, pros, seq = read_data(args_srna)
    out = open(args_srna.output_file, "w")
    out.write("##gff-version 3\n")
    out_t = open(args_srna.output_table, "w")
    get_terminal(cdss, inters, seq, "start")
    get_inter(cdss, inters)
    get_terminal(cdss, inters, seq, "end")
    inters = sorted(inters, key=lambda k: (k["strain"], k["start"],
                                           k["end"], k["strand"]))
    args_srna = ArgsContainer().extend_utr_container(
                            args_srna, cdss, tsss, pros, out,
                            out_t, texs)
    for inter in inters:
        for ta in tas:
            if (inter["strain"] == ta.seq_id) and (
                    inter["strand"] == ta.strand):
                class_utr(inter, ta, args_srna, wig_fs, wig_rs)
    covers = get_utr_coverage(args_srna.utrs)
    mediandict = set_cutoff(covers, args_srna)
    print_median(args_srna.out_folder, mediandict)
    detect_srna(mediandict, args_srna)
    args_srna.out.close()
    args_srna.out_t.close()
    paras = [args_srna.srnas, args_srna.utrs, seq, inters,
             tas, cdss, tas, tsss, pros, covers]
    free_memory(paras)
